;FASM STANDARD LIBRARY
;Linux filesystem routines

;contains:
; proc  file.open       name, mode
; proc  file.create     name, mode
; proc  file.close      handle
; proc  file.seek       handle, dist, method
; proc  file.read       handle, buffer, count
; proc  file.write      handle, buffer, count





;===============================================================================
;file.open
;desc: opens file
;args: name - pointer to null terminated string containing file name
;      mode - 0: read, 1: write, 2:r/w
;ret: CF on error, otherwise
;     EAX = handle to file
;note: don't forget to file.close returned handle!
;error: ERR_FILE_NOT_FOUND - file not found
;       ERR_CANT_OPEN_FILE - file exists, but we failed to open it
;       ERR_INVALID_MODE - invalid mode
;       ERR_UNKNOWN
;===============================================================================
proc file.open name, mode
	push	ebx ecx edx

	cmp	[mode], 2
	ja	.error_invalid_mode

	mov	edx, 0		    ;file permissions
	mov	ecx, [mode]	    ;file access bits (same as our modes)
	mov	ebx, [name]	    ;file name pointer
	mov	eax, 5		    ;sys_open
	int	80h

	cmp	eax, -4096
	ja	.error_open

	clc

.r:	pop	edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_open:

	;check file not found
	push	ERR_FILE_NOT_FOUND
	cmp	eax, -2 ;ENOENT, ENOFILE
	je	@f
	mov	[esp], eax
@@:	pop	eax
	je	.rc

	;check can't open
	push	ERR_CANT_OPEN_FILE
	cmp	eax, -13 ;EACCES
	je	@f
	cmp	eax, -14 ;EFAULT
	je	@f
	cmp	eax, -21 ;EISDIR
	je	@f
	cmp	eax, -26 ;ETXTBUSY
	je	@f
	cmp	eax, -28 ;ENOSPC
	je	@f
	cmp	eax, -30 ;EROFS
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;===============================================================================
;file.create
;desc: creates new file
;args: name - pointer to null terminated string caontaining file name
;      mode - bit0: 1 if overwrite when file exists
;             bit1: 1 if you want read access too
;             all other bits 0
;ret: CF on error, otherwise
;     EAX = handle to file
;note: don't forget to file.close returned handle!
;error: ERR_FILE_EXISTS        - file already exists
;       ERR_CANT_OPEN_FILE     - file can't be created
;       ERR_INVALID_MODE       - if mode is invalid
;       ERR_UNKNOWN
;===============================================================================
proc file.create name, mode
	push ebx ecx edx

	;check if mode is valid
	cmp	[mode], 4
	jae	.error_invalid_mode

	;ECX = file access bits, depending on [mode]
	mov	ecx, 80h+41h		;O_EXCL | O_CREAT | O_WRONLY
	test	[mode], 1b
	jz	@f
	xor	ecx, 80h or 200h	;O_EXCL -> O_TRUNC if bit0 set
@@:	test	[mode], 10b
	jz	@f
	inc	ecx			;O_WRONLY->O_RDWR if bit1 set
@@:

	mov	edx, 110100100b 	;file permissions (644)
					;ecx = file access bits already
	mov	ebx, [name]		;ebx = file name
	mov	eax, 5			;sys_open
	int	80h
	cmp	eax, -4096
	ja	.error_create

	clc

.r:	pop	edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_create:

	;catch already-exists
	push	ERR_FILE_EXISTS
	cmp	eax, -17 ;EEXIST
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	;catch can't open
	push	ERR_CANT_OPEN_FILE
	cmp	eax, -2 ;ENOENT
	je	@f
	cmp	eax, -13 ;EACCES
	je	@f
	cmp	eax, -14 ;EFAULT
	je	@f
	cmp	eax, -21 ;EISDIR
	je	@f
	cmp	eax, -26 ;ETXTBUSY
	je	@f
	cmp	eax, -28 ;ENOSPC
	je	@f
	cmp	eax, -30 ;EROFS
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc


.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;===============================================================================
;file.close
;desc: closes file handle
;args: handle - file handle
;ret: CF set on error
;error: ERR_INVALID_HANDLE - wrong handle
;       ERR_UNKNOWN
;===============================================================================
proc file.close handle
	push	ebx

	mov	ebx, [handle]
	mov	eax, 6
	int	80h
	cmp	eax, -4096
	ja	.error_close

	clc

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_close:
	;catch EBADF
	cmp	eax, -9; EBADF
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
@@:
	;others
	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;===============================================================================
;file.seek
;desc: seek position in file
;args: handle - file handle (returned by file.open)
;      dist   - signed distance to move
;      mdoe   - how to seek (0: from beginning, 1:relative to current, 2: from end)
;ret: CF set on error, otherwise
;     EAX = new file position (from beginning)
;note: - you can use this to get current position
;          libcall file.seek, <handle>, 0, 1
;      - you can use this to get size of opened file:
;          libcall file.seek, <handle>, 0, 2
;        but note that this also changes position in file to end
;      - you can safely use this call only on files smaller than 2GB
;error: ERR_INVALID_HANDLE - invalid file handle
;       ERR_OUT_OF_RANGE   - dist<0 when method=0, or seek to negative file offset
;       ERR_INVALID_MODE   - method is not 0, 1, or 2
;       ERR_UNKNOWN
;===============================================================================
proc file.seek handle, dist, mode
	push	ebx ecx edx

	;check if mode is 0, 1, or 2
	cmp	[mode], 2
	jbe	@f
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
@@:

	;check if dist is positive if mode is 0
	cmp	[mode], 0
	jne	@f
	cmp	[dist], 0
	jge	@f
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
@@:

	mov	edx, [mode]
	mov	ecx, [dist]
	mov	ebx, [handle]
	mov	eax, 19 	;sys_seek
	int	80h
	cmp	eax, -4096
	ja	.error_seek

.rnc:	clc
.r:	pop	edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_seek:
	;catch out-of-bounds
	cmp	eax, -22; EINVAL
	jne	@f
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
@@:

	;catch invalid-handle
	cmp	eax, -9; EBADF
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
@@:

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;===============================================================================
;file.read
;desc: read data from current position in file
;args: handle - file handle (returned by file.open)
;      buffer - pointer to buffer where to store data
;      count - number of bytes to read
;ret: CF set on error, otherwise
;     OF set on read behind EOF (less that <count> bytes were readen)
;     EAX = number of bytes readen
;     ZF set if no bytes were readen (we already were at/behind EOF), then EAX=0
;error: ERR_INVALID_HANDLE - invalid handle, or handle is write-only
;       ERR_ZERO_SIZE      - count is zero
;       ERR_UNKNOWN
;===============================================================================
proc file.read handle, buffer, count
	push	ebx ecx edx
	pushf

	;error if count=0
	cmp	[count],0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:

	;do the read
	mov	edx, [count]
	mov	ecx, [buffer]
	mov	ebx, [handle]
	mov	eax, 3		;sys_read
	int	80h
	cmp	eax, -4096
	ja	.error_read

	;no error, clear CF
	and	byte [esp],not 1		;clear CF

	;set OF if bytes readen != bytes to read
	and	dword [esp],not (1 shl 11)	;clear OF
	cmp	eax,[count]
	je	@f
	or	dword [esp],1 shl 11		;set OF
@@:

	;set ZF if no bytes were readen
	and	byte [esp],not (1 shl 6)	;clear ZF
	cmp	eax, 0
	ja	@f
	or	byte [esp], 1 shl 6		;set ZF
@@:

	;return number of bytes readen in EAX

.r:	popf
	pop	edx ecx ebx
	ret

	;error occured
.rc:	or	dword [esp], 1 ;set CF
	jmp .r


.error_read:
	;catch EBADF
	cmp	eax, -9; EBADF
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	stc
	jmp	.rc
@@:

	mov	eax, ERR_UNKNOWN
	jmp	.rc

endp


;===============================================================================
;file.write
;desc: writes data to current position in file
;args: handle - file handle (returned by file.open)
;      buffer - pointer to buffer from where to read data
;      size - number of bytes to read
;ret: CF set on error
;error: ERR_INVALID_HANDLE - invalid handle
;       ERR_ZERO_SIZE      - count is zero
;       ERR_UNKNOWN
;===============================================================================
proc file.write handle, buffer, count
	push	ebx ecx edx

	;error if count=0
	cmp	[count],0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:

	;do the write
	mov	edx, [count]
	mov	ecx, [buffer]
	mov	ebx, [handle]
	mov	eax, 4		;sys_write
	int	80h
	cmp	eax, -4096
	ja	.error_write

	;error if everything wasn't written
	cmp	eax, [count]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
@@:

	;no error, clear CF
	clc
.r:	pop	edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_write:
	;catch EBADF
	cmp	eax, -9; EBADF
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
@@:

	mov	eax, ERR_UNKNOWN
	jmp	.r

endp



;by vid