;FASM STANDARD LIBRARY
;Linux memory routines for use with Heap Manager

;contains:
; proc  mem.alloc_heap       size
; proc  mem.free_heap        heapptr, size
; proc  mem.realloc_heap     heapptr, newsize, oldsize



NEED_HEAP_MANAGER equ 1

;============================================================================
; mem.alloc_heap
; desc: allocates block of memory (heap) of given size
; args: size - size of memory block to allocate
; ret: CF set on error, otherwise
;      eax = pointer to beginning of allocated memory block
; note: - returned pointer must be always >= 10000h
;       - contents of allocated memory are unknown
;         (use alloc0 to have block zeroed)
;       - don't forget to mem.free allocated memory!
;       - maximal allowed size if 7FFFFFFFh, to catch errors with negative value
; error: ERR_OUT_OF_MEM   - out of memory
;        ERR_ZERO_SIZE    - size=0
;        ERR_OUT_OF_RANGE - size >= 80000000h
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;============================================================================
proc mem.alloc_heap  size
	push	ebx ecx

	;error if size=0
	cmp	[size], 0
	je	.error_size_zero

	;error is size>=80000000h
	cmp	[size], 80000000h
	je	.error_size_too_big

	;call brk(0) to find beginning of heap
	mov	ebx, 0
	mov	eax, 45 	;sys_brk
	int	80h
	cmp	eax, -4096
	ja	.error_brk
	mov	ecx, eax

	;call brk(ebx+size) to set up heap size
	mov	ebx, ecx
	add	ebx, [size]	;
	mov	eax, 45 	;sys_brk
	int	80h
	cmp	eax, -4096
	ja	.error_brk

	;return address of heap
	mov	eax, ecx

.rnc:	clc
.r:	pop	ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_brk:

	;catch out-of-memory
	cmp	eax, -12 ;ENOMEM
	jne	@f
	mov	eax, ERR_OUT_OF_MEM
	jmp	.rc
@@:

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_size_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

endp



;============================================================================
; mem.free_heap
; desc: deallocates memory block (heap)
; args: heapptr - pointer to memory block returned by mem.alloc_heap
;       size    - size of block (the one that was passed to mem.alloc_heap)
; ret: CF set on error
; error: ERR_INVALID_HANDLE - if blockptr isn't valid block pointer
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;============================================================================
proc mem.free_heap blockptr, size
	push	ebx ecx

	;call sys_brk(blockptr)
	mov	ebx, [blockptr] ;pointer to block
	mov	eax, 45 	;sys_brk
	int	80h

	;check error
	cmp	eax, -4096
	ja	.error_brk

	clc

.r:	pop	ecx ebx
	ret

.error_brk:

	;catch invalid block pointer
	cmp eax, -22 ;EINVAL
	jne @f
	mov eax, ERR_INVALID_HANDLE
	stc
	jmp .r
@@:
	mov eax, ERR_UNKNOWN
	stc
	jmp	.r

.error_not_initialized:
	mov eax, ERR_MODULE_NOT_INITIALIZED
	stc
	jmp .r

endp



;============================================================================
; mem.realloc
; desc: reallocates memory block
; args: blockptr - pointer to memory block returned by mem.alloc
;       newsize - new size of memory block
; ret: CF set on error, orherwise
;      eax = pointer to reallocated memory block
; note: - returned pointer must be always >= 10000h
;       - data in block is not modified (only data at the end of block are
;         deleted if newsize < previous size)
;       - if block is enlarging, then contents of added memory are unknown
;         (use realloc0 to have it zeroed)
;       - block can move to other place in memory, so previous pointers into
;         block become invalid (so rather use indexes in block than pointers)
;       - if you had pointer stored in variable, don't forget to restore it:
;           libcall mem.realloc, [blockptr], [newsize]
;           jc error
;           mov [blockptr], eax     ;without this, it's hard-to-find bug
; error: ERR_OUT_OF_MEM     - kernel doesn't like you
;        ERR_INVALID_HANDLE - if blockptr isn't valid block pointer
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;============================================================================
proc mem.realloc_heap heapptr, newsize, oldsize
	push	ebx

	;call sys_brk(blockptr)
	mov	ebx, [heapptr] ;pointer to block
	add	ebx, [newsize]
	mov	eax, 45 	;sys_brk
	int	80h

	;check error
	cmp	eax, -4096
	ja	.error_brk

.rnc:	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_brk:
	;catch invalid block pointer
	cmp	eax, -22 ;EINVAL
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
@@:

	;catch out-of-memory
	cmp	eax, -12 ;ENOMEM
	jne	@f
	mov	eax, ERR_OUT_OF_MEM
	jmp	.rc
@@:

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_newsize_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp

;by vid