;FASM STANDARD LIBRARY
;memory routines

;contains:                      @ = platform-specific
;@ proc  mem.init
;@ proc  mem.uninit

;@ proc  mem.alloc       size
;  proc  mem.alloc0      size
;@ proc  mem.free        blockptr
;@ proc  mem.realloc     blockptr, newsize
;  proc  mem.realloc0    blockptr, newsize
;@ proc  mem.size        blockptr

;  proc  mem.fill        dest, value, size
;  proc  mem.copy        dest, src, size
;  proc  mem.cmp         block1, block2, size
;  proc  mem.find        block, blocksize, pattern, patternsize
;  proc  mem.findchar    block, blocksize, value


;include platform-specific part
match os,SYSTEM {
  include `os#'/mem.inc'
}


;=========================================================
;mem.alloc0
;desc: allocates <size> bytes of memory and zeroes them
;args: size - size of memory to allocate
;ret: CF set on error, otherwise
;     EAX = pointer to beginning of allocated memory block
;note: returned pointer must be always >= 10000h
;error: ERR_ZERO_SIZE - if size = 0
;       +mem.alloc
;       +mem.fill
proc mem.alloc0 size

	push	ebx

	;error if size=0
	cmp	[size], 0
	je	.error_size_zero

	;allocate
	libcall mem.alloc, [size]
	jc	.rc
	mov	ebx, eax

	;zero memory
	libcall mem.fill, ebx, [size], 0
	jc	.rc

	mov	eax, ebx

.rnc:	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;=========================================================
;mem.realloc0
;desc: reallocates memory block, and if block grows, zeroes added bytes
;args: blockptr - pointer to memory block returned by mem.alloc
;      newsize - new size of memory block
;ret: CF set on error, orherwise
;     eax = pointer to reallocated memory block
;note: - returned pointer must be always >= 10000h
;      - data in block are not modified (only data at the end of block are
;        deleted if newsize < previous size)
;      - block can move to other place in memory, so previous pointers into
;        block become invalid (so rather use indexes in block than pointers)
;      - if you had pointer stored in variable, don't forget to restore it:
;          libcall mem.realloc0, [blockptr], [newsize]
;          jc error
;          mov [blockptr], eax     ;without this, it's hard-to-find bug
;error: ERR_ZERO_SIZE - if newsize = 0
;       +mem.size
;       +mem.realloc
;       +mem.fill
proc mem.realloc0 blockptr, newsize
	push	ebx esi

	;error if newsize=0
	cmp	[newsize], 0
	je	.error_newsize_zero

	;ESI = save old block size
	libcall mem.size, [blockptr]
	mov	esi, eax

	;reallocate
	libcall mem.realloc, [blockptr], [newsize]
	jc	.r
	mov	ebx, eax

	;number of bytes to zero = newsize-oldsize
	mov	eax, [newsize]
	sub	eax, esi

	;zero new memory
	libcall mem.fill, addr ebx+esi, eax, 0
	jc	.r

	mov	eax, ebx

.rnc:	clc
.r:	pop	esi ebx
	ret
.rc:	stc
	jmp	.r

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

endp



;=========================================================
; mem.fill
; desc: fills memory with constant value
; args: dest - pointer to beginning of memory block to fill
;       value - dword "filler" value
;       size - number of bytes to fill
; ret: CF = 1 on error, otherwise
;      EAX = dest (for use with nested libcalls)
; note: - module doesn't have to be initalized for this to work
;       - use dword operator to push 4byte-string as filler value (libcall mem.fill, [buf], dword "abcd", [size])
;       - does nothing when size=0
; error: none
proc mem.fill, dest, size, value
	push	ecx edi
	pushf
	cld

	mov	edi, [dest]
	mov	ecx, [size]
	mov	eax, [value]

	;just return if size=0
	cmp	ecx,0
	je	.done

	;align filling pointer to dword
	test	edi, 1b
	jz	@f
	stosb
	ror	eax, 8
	dec	ecx
@@:	test	edi, 10b
	jz	@f
	stosw
	ror	eax, 16
	dec	ecx
	dec	ecx
@@:
	;do the main fill
	push	ecx
	shr	ecx, 2
	rep	stosd
	pop	ecx

	;fill rest
	shr	ecx, 1
	jnc	@f
	stosb
	ror	eax, 8
@@:	shr	ecx, 1
	jnc	@f
	stosw
@@:
	;everything is okay
.done:	and	byte [esp],not 1  ;clear CF in stack
	mov	eax, [dest]

.r:	popf
	pop	edi ecx
	ret
endp



;=============================================================================
; mem.copy
; desc: copies block of memory
; args: dest - pointer to beginning of memory block to where to copy
;       src - pointer to beginning of memory block from where to copy
;       size - number of bytes to copy
; ret: CF = 1 on error, otherwise
;      EAX = dest (for use with nested libcalls)
; note: - module doesn't have to be initalized for this to work
;       - does nothing when size=0
;       - handles case when copying from backwards is needed, eg.
;         when src > dest > src+len
; error: none
;=============================================================================
proc mem.copy, .dest, .src, .size
	push	ecx esi edi
	pushf

	mov	edi, [.dest]
	mov	esi, [.src]
	mov	ecx, [.size]

	;just return when copying 0 bytes
	cmp	ecx, 0
	je	.done

	;check if dest area is overlapping src (when src < dest < src+size)
	cmp	esi, edi
	je	.done ;if dest=src, we don't need to move
	ja	.not_overlapping
	lea	eax, [esi+ecx]
	cmp	edi, eax
	jae	.not_overlapping


	;if it's overlapping
.overlapping:
	;start moving from end of buffers
	add	esi, ecx
	add	edi, ecx

	;classical unoptimized unalgined move
	std
	dec	esi
	dec	edi
	shr	ecx, 1
	jnc	@f
	movsb
@@:

	dec	esi
	dec	edi
	shr	ecx, 1
	jnc	@f
	movsw
@@:	jecxz	@f

	sub	esi, 2
	sub	edi, 2
	rep	movsd
@@:

	jmp	.done


.not_overlapping:

	;classical unoptimized unalgined move
	cld
	shr	ecx, 1
	jnc	@f
	movsb
@@:	shr	ecx, 1
	jnc	@f
	movsw
@@:	jecxz	@f
	rep	movsd
@@:
	;everything is okay
.done:	and	byte [esp],not 1     ;clear CF in stack
	mov	eax,[.dest]

.r:	popf
	pop	edi esi ecx
	ret
endp



;=========================================================
; mem.cmp
; desc: compares blocks of memory
; args: block1 - pointer to beginning of first memory block
;       block2 - pointer to beginning of second memory block
;       size   - number of bytes to compare
; ret: CF = 1 on error, otherwise
;      flags set for signed conditional jumps (je, jne, jg, jge, jl, jle and equivalents), and
;      if ZF=0, eax = offset in blocks (not pointer) where blocks differ
; note: - module doesn't have to be initalized for this to work
;       - when ZF=1 (eg. when blocks are equal), EAX=size
; error: ERR_ZERO_SIZE - size=0
proc mem.cmp block1, block2, size
	push	esi edi ecx
	pushf
	cld

	mov	edi, [block1]
	mov	esi, [block2]
	mov	ecx, [size]

	;error if size=0
	cmp	ecx, 0
	je	.error_size_zero

	;do the compare
	repe	cmpsb
	je	.equal

	;set flags to greater/lesser
	mov	al, byte [edi-1]
	cmp	al, byte [esi-1]
	ja	.greater

	;flags for conditional jumps
	; je:      ZF=1
	; jne:     ZF=0
	; jg/jnle: ZF=0 & (SF=OF)
	; jge/jnl: SF=OF
	; jl/jnge: SF!=OF
	; jle/jng: ZF=1 | (SF!=OF)
.lesser:
	;ZF=0, SF=0, OF=1
	and	word [esp], not ((1 shl 6) + (1 shl 7)) ;clear ZF and SF
	or	word [esp], 1 shl 11			;set OF
	jmp	.done

.greater:
	;ZF=0, SF=1, OF=1
	and	word [esp], not (1 shl 6)		;clear ZF
	or	word [esp], (1 shl 11) + (1 shl 7)	;set OF and SF
	jmp	.done

.equal:
	;ZF=1, SF=1, OF=1
	or	word [esp], (1 shl 6) + (1 shl 7) + (1 shl 11)

.done:	;EAX = offset where blocks differ
	mov	eax, [size]
	sub	eax, ecx
	dec	eax

.rnc:	and	byte [esp],not 1     ;clear CF
.r:	popf
	pop	ecx edi esi
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;=========================================================
; mem.find
; desc: find pattern in block of memory
; args: block - pointer to beginning memory block where to search
;       blocksize - size of memory block in which we search
;       pattern - pointer to beginning of pattern which we are searching for
;       patternsize - it's size
; ret: CF = 1 on error, otherwise
;      ZF = 1 if not found, otherwise
;      eax = offset (not pointer) where the pattern was found,
;            in range 0 to block+blocksize-patternsize
; note: - module doesn't have to be initalized for this to work
;       - uses Horspool's algorithm (http://www-igm.univ-mlv.fr/~lecroq/string/node18.html)
; error: ERR_ZERO_SIZE - patternsize = 0
;=============================================================================
proc mem.find  block, blocksize, pattern, patternsize
local shift_table rb 256*4

	push	esi edi ecx edx ebx
	pushf
	cld

	;error if patternsize = 0
	cmp	[patternsize], 0
	je	.error_patternsize_zero

	;if patternsize>blocksize return nothing found
	mov	eax, [patternsize]
	cmp	eax, [blocksize]
	ja	.not_found


	;fill every shift_table memebr with patternsize
	libcall mem.fill, addr shift_table, 1024, [patternsize]
	jc	.rc

	;fill table with characters from pattern
	mov	ecx, [patternsize]
	dec	ecx
	jz	@f			;if patternsize=1, we don't need to fill table anymore
	lea	edx, [shift_table]
	mov	esi, [pattern]
.fill_table:
	movzx	eax, byte [esi]
	shl	eax, 2
	mov	dword [edx+eax], ecx
	inc	esi
	loop	.fill_table
@@:

	;go with ESI in block through blocksize-patternsize characters
	mov	esi, [block]		;esi = block
	mov	edi, [pattern]		;edi = pattern
	mov	ecx, esi
	add	ecx, [blocksize]
	sub	ecx, [patternsize]	;ecx = place where to stop searching
	lea	ebx, [shift_table]	;ebx = shift_table
.loop:

.check_match:
	;compare last character, and save it in DL
	mov	eax, [patternsize]
	mov	dl, byte [esi+eax-1]
	cmp	dl, byte [edi+eax-1]
	jne	.not_match

	;compare rest of pattern
	mov	eax, [patternsize]
	dec	eax
	libcall mem.cmp, esi, edi, [patternsize]
	jc	.r
	je	.found

	;we didn't find match, shift forward by number from shift_table[dl]
.not_match:
	movzx	eax, dl
	add	esi, [ebx + 4*eax]

	;continue loop
	cmp	esi, ecx
	jbe	.loop

	;nothing found, return ZF
.not_found:
	or	byte [esp], 1 shl 6 ; set ZF
	mov	eax, -1 ;for case someone forgets to check ZF
	jmp	.rnc

	;pattern found - return it's index in EAX, and ZF=0
.found: and	byte [esp], not (1 shl 6) ;clear ZF
	mov	eax, esi
	sub	eax, [block]

.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	ebx edx ecx edi esi
	ret 

.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_patternsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp


;=============================================================================
; mem.find
; desc: finds byte in block of memory.
; args: block     - pointer to beginning memory block where to search
;       blocksize - size of memory block in which we search
;       value     - low order byte of this dword is value we are searching for
; ret: CF = 1 on error, otherwise
;      ZF = 1 if not found, otherwise
;      eax = offset (not pointer) where the pattern was found,
;            in range 0 to block+blocksize
; note: - module doesn't have to be initalized for this to work
;       - char is zero-extended to dword. use "dword 'c'" to push such char
;         with libcall
; error: none
;=============================================================================
proc mem.findchar block, blocksize, value
	push	ecx edi
	pushf

	;if blocksize=0, nothing found
	cmp	[blocksize], 0
	je	.not_found

	;search
	mov	ecx, [blocksize]
	mov	edi, [block]
	mov	al, byte [value]
	cld
	repne	scasb
	jne	.not_found

	;found
	mov	eax, [blocksize]
	sub	eax, ecx
	dec	eax
	jmp	.rnz

	;not found
.not_found:
	mov	eax, -1 ;for case someone forgets to check ZF
	jmp	.rz

.rnz:	and	byte [esp], not (1 shl 6) ;clear ZF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi ecx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.rz:	or	byte [esp], 1 shl 6 ;set ZF
	jmp	.rnc
endp



;=============================================================================
; include heap manager if requested
; value "NEED_HEAP_MANAGER" should be set in platform-dependant part of mem module
match =1, NEED_HEAP_MANAGER
{
  include 'heapmgr.inc'
}
restore NEED_HEAP_MANAGER


;by vid
;mrpink helped with mem.find