;FASM STANDARD LIBRARY
;Static string library



; proc  str.copy        dest, destlen, src, srclen
; proc  str.cat         dest, destlen, src, srclen
; proc  str.catchar     string, buflen, char
; proc  str.ins         dest, destlen, src, srclen, pos
; proc  str.inschar     dest, destlen, char, pos
; proc  str.sub         dest, destlen, src, srclen, index, len
; proc  str.lowcase     string, buflen
; proc  str.upcase      string, buflen

; proc  str.len         string, buflen
; proc  str.cmp         str1, str1len, str2, str2len
; proc  str.cmpI        str1, str1len, str2, str2len
; proc  str.pos         string, buflen, pattern, patlen
; proc  str.poschar     string, buflen, char


;TODO: str.paste


;===============================================================================================
; str.copy
; desc: copy contents of string into another one
; args: dest    - destination buffer pointer
;       destlen - length of destination buffer
;       src     - pointer to string whose contents will be copied to dest
;       srclen  - length of source string buffer
; ret: CF set on error, otherwise
;      EAX = dest (for nested libcalls)
; error: ERR_ZERO_SIZE   - destlen=0 or srclen=0
;        ERR_BUFFER_FULL - destination buffer not long enough. in this case only first destlen chars
;                          are copied, and dest buffer is not terminated by zero
;===============================================================================================
proc str.copy dest, destlen, src, srclen
	push	edi esi ecx ebx
	cld

	;EBX = destlen, error if destlen=0
	mov	ebx, [destlen]
	cmp	ebx, 0
	je	.error_destlen_zero

	;ECX = srclen, error if srclen=0
	mov	ecx, [srclen]
	cmp	ecx, 0
	je	.error_srclen_zero

	;copy at most max(EBX,ECX) characters, until ending zero
	mov	esi, [src]
	mov	edi, [dest]
.copy_char:
	lodsb
	stosb
	cmp	al, 0
	je	.done
	dec	ebx		;this test must be before src overflow test, so we can always
	jz	.dest_overflow	;add ending zero in .source_overflow
	dec	ecx
	jnz	.copy_char

	;end of source string buffer reached, without finding ending zero
.source_overflow:
	mov	byte [edi], 0		;add ending zero to dest string
	jmp	.done

	;end of destination buffer reached, before storing ending zero
.dest_overflow:
	mov	eax, ERR_BUFFER_FULL
	stc
	jmp	.r

	;done, eax=dest
.done:	mov	eax, [dest]
	clc

.r:	pop	ebx ecx esi edi
	ret

.error_destlen_zero:
	mov	eax, ERR_ZERO_SIZE
	stc
	jmp	.r

.error_srclen_zero:
	mov	eax, ERR_ZERO_SIZE
	stc
	jmp	.r
endp



;==============================================================================================;
; str.cat
; desc: append one string to another
; args: dest    - pointer to destination string buffer
;       destlen - size of destination string buffer
;       src     - pointer to source string
;       srclen  - size of source string buffer
; ret: CF set on error, otherwise
;      EAX = dest
; error: ERR_ZERO_SIZE   - destlen=0 or srclen=0
;        ERR_BUFFER_FULL - destination buffer not long enough. in this case only first destlen chars
;                          are copied, and dest buffer is not terminated by zero
;        +str.len
;        +str.copy
;==============================================================================================;
proc str.cat dest, destlen, src, srclen
	push	ecx

	;error if destlen=0
	cmp	[destlen], 0
	je	.error_destlen_zero

	;error if srclen=0
	cmp	[srclen], 0
	je	.error_srclen_zero

	;get length of dest string
	libcall str.len, [dest], [destlen]
	jc	.rc

	;if dest string doesn't contain ending zero, return with OF=1
	jo	.error_dest_buffer_full

	;ECX = remaining size of dest buffer
	mov	ecx, [destlen]
	sub	ecx, eax

	;EAX = place where to start appending
	add	eax, [dest]

	;copy
	libcall str.copy, eax, ecx, [src], [srclen]
	jc	.rc

	mov	eax, [dest]

.rnc:	clc
.r:	pop	ecx
	ret
.rc:	stc
	jmp	.r


.error_dest_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_srclen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

endp



;==============================================================================================;
; str.catchar
; desc: append character to end of string
; args: string - pointer to string to which the char will be appended
;       buflen - length of string buffer
;       char   - character which will be appended, low-order byte of dword used
; ret: CF set on error, otherwise
;      EAX = string, and
; note: char arument is stored as dword, use "dword 'c'" to push such char with libcall macro
; error: ERR_ZERO_SIZE   - buflen=0
;        ERR_ZERO_VALUE  - char=0
;        ERR_BUFFER_FULL - destination buffer not long enough. in this case only first destlen chars
;                          are copied, and dest buffer is not terminated by zero
;        +str.len
;==============================================================================================;
proc str.catchar string, buflen, char
	push	edi ecx

	;error if buflen=0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;error if char=0
	cmp	[char], 0
	je	.error_char_zero

	;EDI = string
	mov	edi, [string]

	;get length of string
	libcall str.len, [string], [buflen]
	jc	.rc
	jo	.error_dest_buffer_full

	;append char
	mov	cl, byte [char]
	mov	byte [edi+eax], cl

	;error if there is no more space for ending zero
	inc	eax
	cmp	eax, [buflen]
	je	.error_dest_buffer_full

	;append ending zero
	mov	byte [edi+eax], 0

	;return string
	mov	eax, edi
	clc

.r:	pop	ecx edi
	ret
.rc:	stc
	jmp	.r

.error_dest_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_char_zero:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc
endp



;==============================================================================================;
; str.ins
; desc: insert string into another string
; args: dest    - pointer to string into which we insert
;       destlen - length of destination string buffer
;       src     - pointer to string which we insert
;       srclen  - length of buffer of string which we insert
;       pos     - position in destination string where we insert
; ret: CF set on error, otherwise
;      EAX = dest
; error: ERR_OUT_OF_RANGE - pos is greater than strlen(dest)
;        ERR_ZERO_SIZE    - destlen=0 or srclen=0
;        ERR_BUFFER_FULL  - destination buffer not long enough. in this case only first destlen chars
;                           of resulting string are in buffer, and dest buffer is not terminated by zero.
;        +str.len
;        +mem.copy
;==============================================================================================;
proc str.ins dest, destlen, src, srclen, pos
	push	ebx ecx esi edi
	cld

	;error if destlen=0
	cmp	[destlen], 0
	je	.error_destlen_zero

	;error if srclen=0
	cmp	[srclen], 0
	je	.error_destlen_zero

	;ECX = length of source string
	libcall str.len, [src], [srclen]
	jc	.rc
	mov	ecx, eax

	;just return when source string is empty
	;??? - returns without ERR_BUFFER_FULL even when dest string is not terminated...
	cmp	ecx, 0
	je	.done

	;EBX = length of dest string, return with OF if not zero terminated
	libcall str.len, [dest], [destlen]
	jc	.rc
	mov	ebx, eax

	;check if pos is inside dest string (pos <= str.len(dest))
	cmp	[pos], ebx
	jbe	@f
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
@@:

	;we don't need to shift characters from dest string if source string
	;fills entire buffer, eg. str.len(src) >= destlen-pos
	mov	eax, [destlen]
	sub	eax, [pos]
	cmp	ecx, eax
	ja	.copy_src

	;we also don't need to shift characters from dest string if we are
	;inserting to it's end
	cmp	[pos], ebx
	je	.copy_src

	;but we have to check case when there is not enough space for moving
	;all of characters, eg str.len(dest) + str.len(src) > destlen
	lea	eax, [ebx+ecx]
	cmp	eax, [destlen]
	jbe	.can_shift_all

	;we can't shift all characters. find how many we can
	;it's: destlen - (pos + str.len(src))
	mov	eax, [destlen]
	sub	eax, [pos]
	sub	eax, ecx
	jmp	.shift_dest

	;we can shift all characters - that means str.len(dest)-pos
.can_shift_all:
	mov	eax, ebx
	sub	eax, [pos]

	;move EAX characters by ECX places forward to make space for new string
.shift_dest:
	mov	esi, [dest]
	add	esi, [pos]    ;from dest[pos]
	mov	edi, esi
	add	edi, ecx      ;to dest[pos+str.len(src)]
	libcall mem.copy, edi, esi, eax
	jc	.rc

	;copy source into destination
.copy_src:

	;EAX = how many bytes we can copy = min(str.len(src), destlen-pos)
	mov	eax, [destlen]
	sub	eax, [pos]
	cmp	eax, ecx
	cmova	eax, ecx

	;error if we are trying to insert at end of string which is also end of buffer
	cmp	eax, 0
	je	.error_dest_buffer_full

	;let us copy
	mov	edi, [dest]
	add	edi, [pos]
	libcall mem.copy, edi, [src], eax

	;error if there isn't space left for ending zero
	lea	eax, [ebx+ecx]
	cmp	eax, [destlen]
	jae	.error_dest_buffer_full

	;now add ending zero to end of string, eg. to: min(str.len(dest)+str.len(src), destlen-1)
	lea	eax, [ebx+ecx]
	mov	ebx, [destlen]
	dec	ebx
	cmp	eax, ebx
	cmova	eax, ebx
	add	eax, [dest]
	mov	byte [eax], 0


.done:	mov	eax,[dest]
	clc
.r:	pop	edi esi ecx ebx
	ret

.rc:	stc
	jmp	.r

.error_dest_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_srclen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;==============================================================================================;
; str.inschar
; desc: insert character into string
; args: string - pointer to destination string
;       buflen - length of destination string's buffer
;       char   - character which will be inserted, low-order byte of dword used
;       pos    - position where to store character
; ret: CF set on error, otherwise
;      EAX = string
; note: char arument is stored as dword, use "dword 'c'" to push such char with libcall macro
; error: ERR_ZERO_VALUE - char=0
;        +str.ins
;==============================================================================================;
proc str.inschar string, buflen, char, pos

	;error if char=0
	cmp	[char], 0
	jne	@f
	mov	eax, ERR_ZERO_VALUE
	stc
	ret
@@:

	;erm.... sorry :)
	libcall str.ins, [string], [buflen], addr char, 1, [pos]
	ret

endp



;===============================================================================================
; str.sub
; desc: extract substring from string
; args: dest    - destination string buffer
;       destlen - destination string buffer length
;       src     - source string pointer
;       srclen  - source string buffer length
;       index   - index in string from where we start extracting
;       len     - number of characters to take into new string
; ret: CF set on error, otherwise
;      EAX = dest (for use with nested libcalls), and
;      OF set if index+len is behind source string (eg. dest string contains less than len chars)
; error: ERR_OUT_OF_RANGE - index is higher than length of string (or length of buffer)
;        ERR_ZERO_SIZE - destlen=0 or srclen=0 or len=0
;        ERR_BUFFER_FULL - destination buffer not long enough. in this case only first destlen chars
;                          are copied, and dest buffer is not terminated by zero
;        +str.len
;        +mem.copy
;===============================================================================================
proc str.sub dest, destlen, src, srclen, index, len
	push	ecx esi
	pushf
	cld

	;error if destlen=0
	cmp	[destlen], 0
	je	.error_destlen_zero

	;error if srclen=0
	cmp	[srclen], 0
	je	.error_srclen_zero

	;error if len=0
	cmp	[len],0
	je	.error_len_zero

	;ECX = length of source string
	libcall str.len, [src], [srclen]
	jc	.rc
	mov	ecx, eax

	;ERR_OUT_OF_RANGE if index >= length of source string
	cmp	[index], ecx
	jae	.error_index_out_of_range

	;ECX = required length for the new string = min(index+len, str.len(string)) - index
	mov	eax, [index]
	add	eax, [len]
	cmp	eax, ecx
	jbe	@f
	mov	eax, ecx
	or	word [esp], 1 shl 11 ;set OF
@@:
	sub	eax, [index]
	mov	ecx, eax

	;ECX = length in dest buffer = min(ECX, destlen)
	mov	eax, [destlen]
	cmp	ecx, eax
	cmova	ecx, eax

	;get position in source string from where to copy
	mov	esi, [src]
	add	esi, [index]

	;copy actual part of string that should be copied
	libcall mem.copy, [dest], esi, ecx
	jc	.rc

	;error if there is no place for ending zero, eg. if ECX = destlen
	cmp	ecx, [destlen]
	je	.error_dest_buffer_full

	;otherwise add ending zero
	add	ecx, [dest]
	mov	byte [ecx], 0

	;return dest
	mov	eax, [dest]

.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	esi ecx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_dest_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_srclen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_index_out_of_range:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

endp



;==============================================================================================;
; str.lowcase
; desc: converts string to lowcase (converts all characters 'A'-'Z' to 'a'-'z')
; args: string - pointer to destination string
;       buflen - length of destination string's buffen
; ret: CF set on error, otherwise
;      EAX = string (for use with nested libcalls)
; error: ERR_ZERO_SIZE - buflen=0
;==============================================================================================;
proc str.lowcase string, buflen
	push	edi ecx

	;ECX = buflen, error if buflen=0
	mov	ecx, [buflen]
	cmp	ecx, 0
	je	.error_buflen_zero

	;EDI = string
	mov	edi, [string]

	;loop for each character
.convert_char:

	;conver character and move EDI to next one
	mov	al, [edi]
	cmp	al, 'A'
	jb	@f
	cmp	al, 'Z'
	ja	@f
	or	byte [edi], 20h
@@:	inc	edi

	;check if we reached end of buffer
	dec	ecx
	jz	.buffer_overflow

	;check if we reached ending zero
	cmp	al, 0
	jnz	.convert_char

	;string buffer wasn't terminated with zero
.buffer_overflow:

.done:	mov	eax, [string]
	clc

.r:	pop	ecx edi
	ret
.rc:	stc
	jmp	.r

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;==============================================================================================;
; str.upcase
; desc: converts string to upcase (converts all characters 'a'-'z' to 'A'-'Z')
; args: string - pointer to destination string
;       buflen - length of destination string's buffen
; ret: CF set on error, otherwise
;      EAX = string (for use with nested libcalls)
; error: ERR_ZERO_SIZE - buflen=0
;==============================================================================================;
proc str.upcase string, buflen
	push	edi ecx

	;ECX = buflen, error if buflen=0
	mov	ecx, [buflen]
	cmp	ecx, 0
	je	.error_buflen_zero

	;EDI = string
	mov	edi, [string]

	;loop for each character
.convert_char:

	;conver character and move EDI to next one
	mov	al, [edi]
	cmp	al, 'a'
	jb	@f
	cmp	al, 'z'
	ja	@f
	and	byte [edi], not 20h
@@:	inc	edi

	;check if we reached end of buffer
	dec	ecx
	jz	.buffer_overflow

	;check if we reached ending zero
	cmp	al, 0
	jnz	.convert_char

	;string buffer wasn't terminated with zero
.buffer_overflow:

.done:	mov	eax, [string]
	clc

.r:	pop	ecx edi
	ret
.rc:	stc
	jmp	.r

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;===============================================================================================
; str.len
; desc: gets length of string
; args: string - string pointer
;       buflen - length of string buffer
; ret: CF set on error, otherwise
;      OF set and EAX=buflen if ending 0 wasn't found in buffer, otherwise
;      OF clear and EAX=length of string
; error: ERR_ZERO_SIZE - buflen=0
;===============================================================================================
proc str.len string, buflen
	push	ecx edi
	pushf
	cld

	;ERR_ZERO_SIZE if buflen=0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;clear OF in stack
	and	word [esp], not (1 shl 11)

	;find ending zero in buffer
	mov	ecx, [buflen]
	mov	edi, [string]
	xor	al,al
	repne	scasb

	;set OF if ending 0 wasn't found
	je	@f
	or	word [esp], 1 shl 11
	dec	ecx		;to disable effect of "dec eax", if there was no ending zero (ugly, i know...)
@@:

	;EAX = number of readen chars
	mov	eax, [buflen]
	sub	eax, ecx
	dec	eax		;ending zero

	;clear CF
	and	byte [esp], not 1

.r:	popf
	pop	edi ecx
	ret

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
.rc:	or	byte [esp], 1	;set CF in stack
	jmp	.r

endp



;==============================================================================================;
; str.cmp
; desc: compare two strings (case sensitive)
; args: str1    - pointer to first string
;       str1len - pointer to first string's buffer length
;       str2    - pointer to second string
;       str1len - pointer to second string's buffer length
; ret: CF set on error, otherwise
;      flags set for signed conditional jumps (je, jne, jg, jge, jl, jle and equivalents), and
;      if ZF=0, then EAX = offset where strings differ
; note: - use returned value in EAX with caution, because it can point behind end of buffer,
;         if some of strings is not zero terminated. for example str.cmp("abcd",4,"abcde",6)
;         returns EAX=4, which is behind end of buffer of first string
;       - if one string is longer than another, then it is considered "greater".
; error: ERR_ZERO_SIZE - str1len=0 or str2len=0
;        +str.len
;==============================================================================================;
proc str.cmp str1, str1len, str2, str2len
	push	ebx ecx edx esi edi
	pushf

	;error if str1len = 0
	cmp	[str1len], 0
	je	.error_str1len_zero

	;error if str2len = 0
	cmp	[str2len], 0
	je	.error_str2len_zero

	;EDI=str1, ESI=str2, EBX=str1len, ECX=str2len
	mov	edi, [str1]
	mov	ebx, [str1len]
	mov	esi, [str2]
	mov	ecx, [str2len]

	;EDX = index of current character = 0
	xor	edx, edx

	;scan each character
.loop:
	;check for end of buffers
	cmp	edx, [str1len]
	je	.str1_end
	cmp	edx, [str2len]
	je	.greater

	;compare characters
	mov	 al, [esi+edx]
	mov	 ah, [edi+edx]
	cmp	 ah, al
	ja	 .greater
	jb	 .lesser

	;if both are zero, then strings are equal
	cmp	 al, 0	 ;we know ah=al already
	je	 .equal

	inc	 edx
	jmp	 .loop


.str1_end:
	;we need to check if both buffers end aren't reached at same time
	cmp	edx, [str2len]
	je	.equal
	jmp	.lesser



	;flags for conditional jumps
	; je:      ZF=1
	; jne:     ZF=0
	; jg/jnle: ZF=0 & (SF=OF)
	; jge/jnl: SF=OF
	; jl/jnge: SF!=OF
	; jle/jng: ZF=1 | (SF!=OF)
.lesser:
	;ZF=0, SF=0, OF=1
	and	word [esp], not ((1 shl 6) + (1 shl 7)) ;clear ZF and SF
	or	word [esp], 1 shl 11			;set OF
	jmp	.done

.greater:
	;ZF=0, SF=1, OF=1
	and	word [esp], not (1 shl 6)		;clear ZF
	or	word [esp], (1 shl 11) + (1 shl 7)	;set OF and SF
	jmp	.done

.equal:
	;ZF=1, SF=1, OF=1
	or	word [esp], (1 shl 6) + (1 shl 7) + (1 shl 11)

.done:	mov	eax, edx
	and	byte [esp],not 1 ;clear CF
.r:	popf
	pop	edi esi edx ecx ebx
	ret

.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_str1len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_str2len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;==============================================================================================;
; str.cmpI
; desc: compare two strings (case insensitive)
; args: str1    - pointer to first string
;       str1len - pointer to first string's buffer length
;       str2    - pointer to second string
;       str1len - pointer to second string's buffer length
; ret: CF set on error, otherwise
;      EAX = offset where strings differ, and
;      flags set for signed conditional jumps (je, jne, jg, jge, jl, jle and equivalents)
; note: - use returned value in EAX with caution, because it can point behind end of buffer,
;         if some of strings is not zero terminated. for example str.cmp("abcd",4,"abcde",6)
;         returns EAX=4, which is behind end of buffer of first string
;       - if one string is longer than another, then it is considered "greater".
; error: ERR_ZERO_SIZE - str1len=0 or str2len=0
;        +str.len
;==============================================================================================;
proc str.cmpi str1, str1len, str2, str2len
	push	ebx ecx edx esi edi
	pushf

	;error if str1len = 0
	cmp	[str1len], 0
	je	.error_str1len_zero

	;error if str2len = 0
	cmp	[str2len], 0
	je	.error_str2len_zero

	;EDI=str1, ESI=str2, EBX=str1len, ECX=str2len
	mov	edi, [str1]
	mov	ebx, [str1len]
	mov	esi, [str2]
	mov	ecx, [str2len]

	;EDX = index of current character = 0
	xor	edx, edx

	;scan each character
.loop:
	;check for end of buffers
	cmp	edx, [str1len]
	je	.str1_end
	cmp	edx, [str2len]
	je	.greater

	;get and convert characters
	mov	 al, [esi+edx]
	cmp	 al, 'A'
	jb	 @f
	cmp	 al, 'Z'
	ja	 @f
	add	 al, 'a'-'A'
	@@:

	mov	 ah, [edi+edx]
	cmp	 ah, 'A'
	jb	 @f
	cmp	 ah, 'Z'
	ja	 @f
	add	 ah, 'a'-'A'
	@@:

	cmp	 ah, al
	ja	 .greater
	jb	 .lesser

	;if both are zero, then strings are equal
	cmp	 al, 0	 ;we know ah=al already
	je	 .equal

	inc	 edx
	jmp	 .loop


.str1_end:
	;we need to check if both buffers end aren't reached at same time
	cmp	edx, [str2len]
	je	.equal
	jmp	.lesser



	;flags for conditional jumps
	; je:      ZF=1
	; jne:     ZF=0
	; jg/jnle: ZF=0 & (SF=OF)
	; jge/jnl: SF=OF
	; jl/jnge: SF!=OF
	; jle/jng: ZF=1 | (SF!=OF)
.lesser:
	;ZF=0, SF=0, OF=1
	and	word [esp], not ((1 shl 6) + (1 shl 7)) ;clear ZF and SF
	or	word [esp], 1 shl 11			;set OF
	jmp	.done

.greater:
	;ZF=0, SF=1, OF=1
	and	word [esp], not (1 shl 6)		;clear ZF
	or	word [esp], (1 shl 11) + (1 shl 7)	;set OF and SF
	jmp	.done

.equal:
	;ZF=1, SF=1, OF=1
	or	word [esp], (1 shl 6) + (1 shl 7) + (1 shl 11)

.done:	mov	eax, edx
	and	byte [esp],not 1 ;clear CF
.r:	popf
	pop	edi esi edx ecx ebx
	ret

.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_str1len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_str2len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp



;==============================================================================================;
; str.pos
; desc: find position of string inside another string
; args: string  - pointer to string inside which we search
;       buflen  - length of buffer of string, inside which we search :)
;       pattern - pointer to pattern string for which we searching
;       patlen  - length of buffer of pattern string
; ret: CF set on error, otherwise
;      ZF set if pattern doesn't occur inside string, otherwise
;      eax = offset of first occurence of pattern inside string
; error: ERR_ZERO_SIZE  - buflen=0 or patlen=0
;        ERR_ZERO_VALUE - pattern is empty string
;        +str.len
;        +mem.find
;==============================================================================================;
proc str.pos	string, buflen, pattern, patlen
	push	ebx
	pushf
	cld

	;error if buflen=0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;error if patlen=0
	cmp	[patlen], 0
	je	.error_patlen_zero

	;ebx = length of the string to search
	libcall str.len, [string], [buflen]
	jc	.rc
	mov	ebx,eax

	;not found if ebx=0
	cmp	ebx, 0
	je	.not_found

	;edx = length of the pattern string
	libcall str.len, [pattern], [patlen]
	jc	.rc

	;error if it is empty
	cmp	eax, 0
	je	.error_pattern_empty

	;use mem.find to find string
	libcall mem.find, [string], ebx, [pattern], eax
	jc	.rc
	jz	.not_found


.found:
	and	dword [esp],not 1	  ;clear CF
	and	dword [esp],not (1 shl 6) ;clear ZF
	jmp	.r

.not_found:
	and	dword [esp],not 1	  ;clear CF
	or	dword [esp],1 shl 6	  ;set ZF

.r:
	popf
	pop	ebx
	ret

.rc:
	or	dword [esp],1		  ;set CF
	jmp	.r

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_patlen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_pattern_empty:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc
endp



;==============================================================================================;
; str.poschar
; desc: find character position in string
; args: string - pointer to string
;       buflen - length of string buffer
;       char   - low order byte of this dword is character to find inside string.
; ret: CF set on error, otherwise
;      ZF set if character doesn't occur inside string, otherwise
;      EAX = offset of first character occurence inside string
; note: char is zero-extended to dword, use "dword 'c'" to push such char with libcall
; error: ERR_ZERO_SIZE  - buflen=0
;        ERR_ZERO_VALUE - char=0
;==============================================================================================;
proc str.poschar string, buflen, char
	push	esi ecx
	pushf	;to return flags

	;ESI = string
	mov	esi, [string]

	;ECX = buflen, error if 0
	mov	ecx, [buflen]
	cmp	ecx, 0
	je	.error_buflen_zero

	;AH = char for which we search, error if 0
	mov	ah, byte [char]
	cmp	ah, 0
	je	.error_char_zero

	;look for character
.search:
	mov	al,[esi]
	inc	esi
	or	al,al
	jz	.not_found

	cmp	al,ah
	je	.found

	dec	ecx
	jnz	.search

.not_found:
	or	dword [esp],(1 shl 6) ;set ZF in stack
	jmp	.rnc

.found:
	and	dword [esp],not (1 shl 6) ;clear ZF
	mov	eax, [buflen]
	sub	eax, ecx

.rnc:	and	dword [esp],not 1 ;clear CF
.r:	popf
	pop	ecx esi
	ret

.rc:	or	dword [esp],1 ;set CF
	jmp	.r

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_char_zero:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

endp

;by vid