;FASM STANDARD LIBRARY
;stream I/O routines

;contains:                      @ = platform-specific
;@ proc  stream.init
;@ proc  stream.uninit
;@ label stream.stdout
;@ label stream.stdin
;@ label stream.stderr
;  proc  stream.create          objbuf, type, size, read, readnb, peek, peeknb, skip, write, writenb, close, readbuf, rbuflen
;  proc  stream.write           stream, buffer, count
;  proc  stream.writechar       stream, char
;  proc  stream.read            stream, buffer, count
;  proc  stream.readchar        stream
;  proc  stream.readnb          stream, buffer, count
;  proc  stream.peek            stream, count, readlen
;  proc  stream.peekchar        stream
;  proc  stream.peeknb          stream, count, readlen
;  proc  stream.skip            stream, count
;  proc  stream.close           stream


; interface of stream object
struc STREAM type,\
      read, peek, skip, write,\
      readnb, peeknb, writenb,\
      flush, close,\
      rbuftype, rbuf, rbufsize,\
      wbuftype, wbuf, wbufsize
{
align 4 			  ;needed more than it may seem ;)
.:
.sign	      dd 'STRM' 	  ;signature of stream
.type	      dd type+0 	  ;type of stream - kind of internal typechecking and RTTI :)
.read	      dd read+0 	  ;pointer to read method
.peek	      dd peek+0 	  ;pointer to peek method
.skip	      dd skip+0 	  ;pointer to skip method
.write	      dd write+0	  ;pointer to write method
.readnb       dd readnb+0	  ;pointer to non-blocking read method
.peeknb       dd peeknb+0	  ;pointer to non-blocking peek method
.writenb      dd write+0	  ;pointer to non-blocking write method
.flush	      dd flush+0
.close	      dd close+0	  ;pointer to close method

.rbuf	   dd rbuf+0	       ;pointer to read buffer
.rbuf.size dd rbufsize+0       ;size of read buffer
.rbuf.pos  dd 0 	       ;index in read buffer, where unread data starts
.rbuf.cnt  dd 0 	       ;number of bytes in read buffer, must be 0 if rbuf.type=0

.wbuf	   dd wbuf+0	       ;pointer to write buffer
.wbuf.size dd wbufsize+0       ;size of write buffer
.wbuf.pos  dd 0 	       ;index in write buffer, where unwritten data starts
.wbuf.cnt  dd 0 	       ;number of bytes in write buffer, must be 0 if wbuf.type=0

.rbuf.type db rbuftype+0       ;0=no buffer, 1=static, 2=dynamic (can be reallocated)
.rbuf.eof  db 0 	       ;1 if EOF is behind data in buffer
.wbuf.type db wbuftype+0       ;0=no buffer, 1=static, 2=dynamic (can be reallocated)

.allocated db 0 	       ;0 if stream is static, 1 if stream is allocated
align 4
}
virtual at 0
  STREAM STREAM
  sizeof.STREAM = $
  match =1, DISPLAY_STRUC_INFO {
    disphex 'size of STREAM object: ', sizeof.STREAM
  }
end virtual

;include platform-specific part
match os,SYSTEM {
  include `os#'/stream.inc'
}


;=============================================================================
; stream.create
; desc: creates stream object. This call allows you to create static objects,
;       or dynamic object with static buffer, or dynamic objects with dynamic
;       buffer.
; args: objbuf  - pointer to buffer to hold stream object. size must be at
;                 least "sizeof.STREAM". set to 0 for dynamic-allocated stream
;       type    - type of stream. it's just dword number you can use in
;                 methods to check if STREAM.type is of right type
;       size    - number of additional bytes to allocate with object. these
;                 can be used to store data for methods (like handles etc.)
;                 must be 0 if buffer != 0
;       read    - pointer to Read method. 0 if not readable stream
;       readnb  - pointer to non-blocking Read method. 0 if not supported
;                 must be 0 if read=0
;       peek    - pointer to Peek method. 0 if not supported (buffering must
;                 be enabled for peeking then). must be 0 if read=0
;       peeknb  - pointer to non-blocking Peek method. 0 if not supported
;                 (then, stream.peeknb only returns data previously peeked)
;                 must be 0 if read=0
;       skip    - pointer to Skip method. 0 if not supported (then, skipping
;                 is automatically emulated using read method).
;                 must be 0 if read=0
;       write   - pointer to Write method. 0 if not writeable stream.
;       writenb - unsupported yet, ignored
;       flush   - pointer to Flush method. 0  if not flushable stream.
;       close   - pointer to Close method. 0 if not closeable stream
;       readbuf - pointer to read buffer, 0 if none, 1 for dynamic buffer
;                 must be 0 if read=0
;       rbuflen - if readbuf=0, ignored
;                 if readbuf=1, initial size of buffer (allocated at creation)
;                 if readbuf=pointer, size of buffer
;                 must be 0 if read=0
;       writebuf - pointer to write buffer, 0 if none, 1 for dynamic buffer
;                 must be 0 if write=0
;       wbuflen - if writebuf=0, ignored
;                 if writebuf=1, initial size of buffer (allocated at creation)
;                 if writebuf=pointer, size of buffer
;                 must be 0 if write=0
; note: - when pointer=0, size>0: additional memory is filled with zeroes
;       - to create completely dynamic stream (like stream.create), set buffer=0,
;         readbuf=1, rbuflen=0, writebuf=1, wbuflen=0
; ret: EAX = headle of created stream
; error: ERR_HANDLE_NOT_READABLE - read=0 and some of peek, skip, readbug, rbuflen is non-zero
;        ERR_OUT_OF_RANGE - size<0
;        +mem.alloc0 (when pointer=0)
;        +mem.alloc (when readbuf=1)
;===============================================================================================
proc stream.create	objbuf, type, size,\
			read, readnb, peek, peeknb, skip, write,\
			flush, close,\
			readbuf, rbuflen, writebuf, wbuflen
	push	ebx

	;if buffer!=0, then size must be 0
	cmp	[objbuf], 0
	je	@f
	cmp	[size], 0
	jne	.error_arguments
	@@:

	;if readbuf=0, then peek, skip, readbuf and rbuflen must be 0
	cmp	[readbuf], 0
	jne	@f
	cmp	[readbuf], 0
	jne	.error_arguments
	cmp	[rbuflen], 0
	jne	.error_arguments
	@@:

	;catch negative size
	cmp	[size], 0
	jl	.error_negative_size

	;EBX = pointer to stream object buffer, allocate it if requested
	mov	ebx, [objbuf]
	cmp	ebx, 0
	jne	@f
	mov	eax, [size]
	add	eax, sizeof.STREAM
	libcall mem.alloc0, eax
	jc	.rc
	mov	ebx, eax
	@@:

	;create object
	mov	[ebx+STREAM.sign], 'STRM'
	mov	eax, [type]
	mov	[ebx+STREAM.type], eax
	mov	eax, [read]
	mov	[ebx+STREAM.read], eax
	mov	eax, [peek]
	mov	[ebx+STREAM.peek], eax
	mov	eax, [skip]
	mov	[ebx+STREAM.skip], eax
	mov	eax, [write]
	mov	[ebx+STREAM.write], eax
	mov	eax, [close]
	mov	[ebx+STREAM.close], eax
	mov	eax, [flush]
	mov	[ebx+STREAM.flush], eax
	mov	eax, [readnb]
	mov	[ebx+STREAM.readnb], eax
	mov	eax, [peeknb]
	mov	[ebx+STREAM.peeknb], eax

	;set "allocated" flag
	mov	eax, [objbuf]
	cmp	eax, 0
	jne	@f
	mov	[ebx+STREAM.allocated], 1
	@@:

	;fill read buffer info, if read method is supported
	cmp	[ebx+STREAM.read], 0
	je	.rbuf_done

	;check if read buffering is disabled
	cmp	[readbuf], 0
	je	.rbuf_done

	;if it's static buffer, just fill
	cmp	[readbuf], 1
	je	@f
	mov	[ebx+STREAM.rbuf.type], 1
	mov	eax, [readbuf]
	mov	[ebx+STREAM.rbuf], eax
	mov	eax, [rbuflen]
	mov	[ebx+STREAM.rbuf.size], eax
	jmp	.rbuf_done
	@@:

	;it's dynamic buffer, fill it
	mov	[ebx+STREAM.rbuf.type], 2
	cmp	[rbuflen], 0
	je	.rbuf_done
	libcall mem.alloc, [rbuflen]
	jc	.rc				;TODO - free memory allocated up to now
	mov	[ebx + STREAM.rbuf], eax
	mov	eax, [rbuflen]
	mov	[ebx+STREAM.rbuf.size], eax

.rbuf_done:

	;check if write buffering is disabled
	cmp	[writebuf], 0
	je	.wbuf_done

	;if it's static buffer, just fill
	cmp	[writebuf], 1
	je	@f
	mov	[ebx+STREAM.wbuf.type], 1
	mov	eax, [writebuf]
	mov	[ebx+STREAM.wbuf], eax
	mov	eax, [wbuflen]
	mov	[ebx+STREAM.wbuf.size], eax
	jmp	.wbuf_done
	@@:

	;it's dynamic buffer, fill it
	mov	[ebx+STREAM.wbuf.type], 2
	cmp	[wbuflen], 0
	je	.wbuf_done
	libcall mem.alloc, [wbuflen]
	jc	.rc				;TODO - free memory allocated up to now
	mov	[ebx + STREAM.wbuf], eax
	mov	eax, [wbuflen]
	mov	[ebx+STREAM.wbuf.size], eax

.wbuf_done:



	mov	eax, ebx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r


.error_negative_size:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_arguments:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc
endp



;===============================================================================================
; stream.write
; desc: writes buffer to stream
; args: stream - handle of stream to write to
;       buffer - buffer from where to write
;       count  - number of bytes to write. does nothing if count=0
; ret: CF set on error
; error: ERR_INVALID_HANDLE      - stream is not a valid stream handle
;        ERR_STREAM_NOT_WRITABLE - stream is not writeable
;        ERR_FASMLIB             - bug in FASMLIB! Please report.
;        +any error returned by stream write method
;===============================================================================================
proc stream.write, stream, buffer, count

	;error if count is zero
	cmp	[count], 0
	je	.rnc

	;EAX = stream object, error if invalid
	mov	eax, [stream]
	cmp	[eax + STREAM.sign], 'STRM'
	jne	.error_not_stream

	;EAX = write method, error if not present
	mov	eax, [stream]
	mov	eax, [eax + STREAM.write]
	cmp	eax, 0
	je	.error_stream_not_writeable

	;call write method
	libcall eax, [stream], [buffer], [count]
	jc	.rc

	;we don't have write buffer yet, so no stream should be using it
	cmp	eax, 0
	ja	.error_write_buffer_unsupported_yet

.rnc:	clc
.r:	ret
.rc:	stc
	jmp	.r

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_writeable:
	mov	eax, ERR_HANDLE_NOT_WRITEABLE
	jmp	.rc

.error_write_buffer_unsupported_yet:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp



;===============================================================================================
; stream.writechar
; desc: writes character to stream
; args: stream - handle of stream to write to
;       char   - low order byte of this dword is character to write
; ret: CF set on error
; note: you must push char as dword. use "dword 'c'"
; error: ERR_INVALID_HANDLE      - stream is not a valid stream handle
;        ERR_STREAM_NOT_WRITABLE - stream is not writeable
;        +stream.write
;===============================================================================================
proc stream.writechar stream, char
	mov	byte [char+1], 0
	lea	eax, [char]
	libcall stream.write, [stream], eax, 1
	ret
endp



;===============================================================================================
; stream.read
; desc: reads to buffer from stream. blocks if there isn't enough data available immediately.
; args: stream - handle of stream from which to read
;       buffer - buffer to which we read
;       count  - number of bytes to read
; ret: CF set on error, otherwise
;      OF set if EOF was reached, and
;      EAX = number of bytes readen (if OF=0, then EAX=count)
; error: ERR_INVALID_HANDLE      - invalid stream handle
;        ERR_STREAM_NOT_READABLE - stream is not readable
;        ERR_ZERO_SIZE           - count=0
;        ERR_HANDLE_CORRUPT      - stream is corrupt, you are overwriting stream object somewhere,
;                                  or it's error in FASMLIB
;        +mem.copy (when reading from buffer)
;        +any error returned by stream read method
;===============================================================================================
proc stream.read, stream, buffer, count
	push	ebx ecx edx edi
	pushf

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx+STREAM.sign], 'STRM'
	jne	.error_not_stream

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;EDI = buffer, ECX = count
	mov	edi, [buffer]
	mov	ecx, [count]

	;if there is no buffering, go directly to read method
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.call_read

	;check if there is some data on buffer
	cmp	[ebx+STREAM.rbuf.cnt], 0
	ja	.read_buffer

	;no data on buffer, so if EOF flag is clear, call read method
	cmp	[ebx+STREAM.rbuf.eof], 0
	je	.call_read

	;no data on buffer, EOF flag set, return EAX=0 and OF=1
	mov	eax, 0
	jmp	.ro

.read_buffer:

	;EDX = bytes to copy from stream buffer = min(count, stream.rbuf.cnt)
	mov	edx, [ebx+STREAM.rbuf.cnt]
	cmp	edx, ecx
	cmova	edx, ecx

	;copy from stream buffer to dest buffer
	mov	eax, [ebx+STREAM.rbuf]
	add	eax, [ebx+STREAM.rbuf.pos]
	libcall mem.copy, edi, eax, edx
	jc	.rc

	;remove data from buffer (set buffer pointer behind them)
	add	[ebx+STREAM.rbuf.pos], edx
	sub	[ebx+STREAM.rbuf.cnt], edx
	jnz	@f
	mov	[ebx+STREAM.rbuf.pos], 0   ;if we emptied the stream buffer, set pointer to beginning
	@@:

	;eax = number of bytes readen
	mov	eax, edx

	;set new destination buffer's position and length
	add	edi, edx
	sub	ecx, edx

	;if we have read enough, return EAX=count, OF=0
	jnz	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;if EOF flag is set in stream buffer, return EAX=readen, OF=1
	cmp	[ebx + STREAM.rbuf.eof], 1
	jne	@f
	mov	eax, edx
	jmp	.ro
	@@:

	;if we didn't read enough, continue reading with read method

.call_read:

	;EAX = read method, error if not present
	mov	eax, [ebx + STREAM.read]
	cmp	eax, 0
	je	.error_stream_not_readable

	;call read method
	libcall eax, [stream], [buffer], ecx
	jc	.rc

	;store size readen. if OF was returned, then mark it on buffer
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	sub	ecx, eax

.done:

	;EAX = size readen from buffer = count-ECX
	mov	eax, [count]
	sub	eax, ecx

	;return OF if EOF was reached
	cmp	[ebx+STREAM.rbuf.eof], 1
	je	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc
endp


;=============================================================================
; stream.readchar
; desc: reads one character from stream
; args: stream - stream to read character from
; ret: CF set on error, otherwise
;      OF set if we are on EOF, otherwise
;      EAX = readen char zero-extended to dword
; error: +stream.read
;=============================================================================
proc stream.readchar stream
local buf rb 1
	lea	eax, [buf]
	libcall stream.read, eax, 1
	jc	.r
	jo	.r
	mov	al, byte [buf]
	clc
.r:	ret
endp



;=============================================================================
; stream.readnb
; desc: reads to buffer from stream. Doesn't block if there isn't enough data
;       immediately, just returns as much as possible.
; args: stream - handle of stream from which to read
;       buffer - buffer to which we read
;       count  - number of bytes to read
; ret: CF set on error, otherwise
;      OF set if EOF was reached, and
;      EAX = number of bytes readen
; note: this is not asynchronous read. This is just simple non-blocking read,
;       it returns data that is available immediately. It doesn't wait for
;       data (or EOF), like stream.read
; error: ERR_INVALID_HANDLE      - invalid stream handle
;        ERR_STREAM_NOT_READABLE - stream is not readable
;        ERR_ZERO_SIZE           - count=0
;        ERR_HANDLE_CORRUPT      - stream is corrupt, you are overwriting stream object somewhere,
;                                  or it's error in FASMLIB
;        +mem.copy (when reading from buffer)
;        +any error returned by stream's readnb method
;===============================================================================================
;TODO - add support for EOF
proc stream.readnb, stream, buffer, count
	push	ebx ecx edx edi
	pushf

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx+STREAM.sign], 'STRM'
	jne	.error_not_stream

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;EDI = buffer, ECX = count
	mov	edi, [buffer]
	mov	ecx, [count]

	;if there is no buffering, go directly to readnb method
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.call_readnb

	;check if there is some data on buffer
	cmp	[ebx+STREAM.rbuf.cnt], 0
	ja	.read_buffer

	;no data on buffer, so if EOF flag is clear, call read method
	cmp	[ebx+STREAM.rbuf.eof], 0
	je	.call_readnb

	;no data on buffer, EOF flag set, return EAX=0 and OF=1
	mov	eax, 0
	jmp	.ro

.read_buffer:

	;EDX = bytes to copy from stream buffer = min(count, stream.rbuf.cnt)
	mov	edx, [ebx+STREAM.rbuf.cnt]
	cmp	edx, ecx
	cmova	edx, ecx

	;copy from stream buffer to dest buffer
	mov	eax, [ebx+STREAM.rbuf]
	add	eax, [ebx+STREAM.rbuf.pos]
	libcall mem.copy, edi, eax, edx
	jc	.rc

	;remove data from buffer (set buffer pointer behind them)
	add	[ebx+STREAM.rbuf.pos], edx
	sub	[ebx+STREAM.rbuf.cnt], edx
	jnz	@f
	mov	[ebx+STREAM.rbuf.pos], 0   ;if we emptied the stream buffer, set pointer to beginning
	@@:

	;eax = number of bytes readen
	mov	eax, edx

	;set new destination buffer's position and length
	add	edi, edx
	sub	ecx, edx

	;if we have read enough, return EAX=count, OF=0
	jnz	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;if EOF flag is set in stream buffer, return EAX=readen, OF=1
	cmp	[ebx + STREAM.rbuf.eof], 1
	jne	@f
	mov	eax, edx
	jmp	.ro
	@@:

	;if we didn't read enough, continue reading with read method

.call_readnb:

	;EAX = read method, return if not present
	mov	eax, [ebx + STREAM.readnb]
	cmp	eax, 0
	je	.done

	;call read method
	libcall eax, [stream], [buffer], ecx
	jc	.rc

	;store size readen. if OF was returned, then mark it on buffer
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	sub	ecx, eax

.done:

	;EAX = size readen from buffer = count-ECX
	mov	eax, [count]
	sub	eax, ecx

	;return OF if EOF was reached
	cmp	[ebx+STREAM.rbuf.eof], 1
	je	.ro


.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc
endp



;===============================================================================================
; stream.peek
; desc: pre-reads data from stream. blocks if there isn't enough bytes available.
;       data remains in stream buffer, and will be returned by stream.read, or stream.skip
; args: stream  - handle of stream from which to read
;       count   - number of bytes to read.
;       readlen - pointer to variable, which will hold number of chars in returned buffer
;                 only needed when function returns OF, otherwise it's equal to count
;                 you can leave this 0, when you don't want to handle case with OF=1
; ret: CF set on error, otherwise
;      OF set if EOF was reached (less than count bytes readen), and
;      EAX = pointer to buffer (READ ONLY, don't write there)
; note: - pointer to buffer becomes invalid after next operation with stream
;       - on OF=0, do not read behind count bytes
;       - on OF=1, do not read buffer behind number of bytes returned in readlen variable.
;         If you passed 0 as readlen, don't read buffer at all
; error: ERR_ZERO_SIZE           - count=0
;        ERR_INVALID_HANDLE      - not a stream handle
;        ERR_HANDLE_NO_BUFFER    - stream doesn't have read buffer associated, nor peek method supported
;        ERR_HANDLE_NOT_READABLE - stream is not readable
;        ERR_BUFFER_FULL         - handle is associated with static buffer, with length smaller than count
;                                  on this error, nothing is read from stream
;        ERR_HANDLE_CORRUPT      - stream is corrupt, you are overwriting stream object somewhere,
;                                  or it's error in FASMLIB (unlikely)
;        +mem.realloc
;        +any error returned by stream read method
;===============================================================================================
proc stream.peek, stream, count, readlen
	push	ebx ecx edi
	pushf

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx+STREAM.sign], 'STRM'
	jne	.error_not_stream

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;error if stream is not readable
	cmp	[ebx+STREAM.read], 0
	je	.error_stream_not_readable

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;first try to use peek method, if provided and nothing is on buffer
	cmp	[ebx+STREAM.rbuf.cnt], 0
	ja	@f
	mov	eax, [ebx+STREAM.peek]
	cmp	eax, 0
	je	@f
	libcall eax, [stream], [count], [readlen]
	jc	.rc ;TODO - we should handle ERR_BUFFER_FULL, and try to buffer at this level, in that case?
	jo	.ro
	jmp	.rno
	@@:

	;ECX = count
	mov	ecx, [count]

	;error if we doesn't have any buffer
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.error_no_buffer

	;error if we have static buffer smaller than count
	cmp	[ebx+STREAM.rbuf.type], 1
	jne	@f
	cmp	ecx, [ebx+STREAM.rbuf.size]
	ja	.error_buffer_small
	@@:

	;if we have dynamic buffer, not yet allocated, then allocate it
	cmp	[ebx+STREAM.rbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.rbuf], 0
	jne	@f
	libcall mem.alloc, ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx
	mov	[ebx+STREAM.rbuf.pos], 0
	mov	[ebx+STREAM.rbuf.cnt], 0
	jmp	.buffer_long_enough
	@@:

	;see if buffer is big enough for count (without moving needed)
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	ecx, eax
	jbe	.buffer_long_enough

	;move data to beginning of buffer
	mov	eax, [ebx+STREAM.rbuf.pos]
	add	eax, [ebx+STREAM.rbuf]
	libcall mem.copy, [ebx+STREAM.rbuf], eax, [ebx+STREAM.rbuf.cnt]
	jc	.rc
	mov	[ebx+STREAM.rbuf.pos], 0

	;see if buffer is long enough now
	cmp	ecx, [ebx+STREAM.rbuf.size]
	jbe	.buffer_long_enough

	;if not, reallocate buffer to needed size
	libcall mem.realloc, [ebx+STREAM.rbuf], ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx

.buffer_long_enough:

	;check if there is already enough bytes in buffer
	cmp	ecx, [ebx+STREAM.rbuf.cnt]
	jbe	.buffer_filled

	;check if there is EOF marked in buffer
	cmp	[ebx+STREAM.rbuf.eof],1
	je	.buffer_filled

	;read remaining bytes
	mov	edi, [ebx+STREAM.rbuf]
	add	edi, [ebx+STREAM.rbuf.pos]
	add	edi, [ebx+STREAM.rbuf.cnt]
	sub	ecx, [ebx+STREAM.rbuf.cnt]
	libcall [ebx + STREAM.read], [stream], edi, ecx
	jc	.rc

	;adjust buffer info (count, EOF flag)
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	add	[ebx+STREAM.rbuf.cnt], eax

.buffer_filled:

	;readlen = min(count, readbuf.cnt)
	mov	ecx, [count]
	mov	eax, [ebx+STREAM.rbuf.cnt]
	cmp	ecx, eax
	cmova	ecx, eax
	mov	eax, [readlen]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;eax = readbuf
	mov	eax, [ebx + STREAM.rbuf]
	add	eax, [ebx + STREAM.rbuf.pos]

	;if user didn't set readlen, and we are returning OF, then return 0 in EAX,
	;so he immediately get GPF if he tries to read from buffer.
	cmp	[readlen], 0
	jne	@f
	cmp	ecx, [count]
	jnb	@f
	mov	eax, 0
	@@:

	;return OF if count<readbuf.cnt
	cmp	ecx, [count]
	jb	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc

.error_no_buffer:
	mov	eax, ERR_HANDLE_NO_BUFFER
	jmp	.rc

.error_buffer_small:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc
endp



;=============================================================================
; stream.peekchar
; desc: peeks one character from stream.
; args: stream - stream to peek character from
; ret: CF set on error, otherwise
;      OF set if we are on EOF, otherwise
;      EAX = peeked char zero-extended to dword
; error: +stream.peek
;=============================================================================
proc stream.peekchar stream
	libcall stream.peek, [stream], 1, 0
	jc	.r
	jo	.r
	movzx	eax, byte [eax]
	cmp	eax, eax		;clear OF
	clc
.r:	ret
endp



;===============================================================================================
; stream.peeknb
; desc: pre-reads data from stream. doesn't block, returns only data available immediately.
;       data remains in stream buffer, and will be returned by stream.read, or stream.skip
; args: stream  - handle of stream from which to read
;       count   - number of bytes to read
;       readlen - pointer to variable, which will hold number of chars in returned buffer
;                 *must* be present, unlike in stream.peek
; ret: CF set on error, otherwise
;      OF set if EOF was reached, and
;      EAX = pointer to buffer (READ ONLY, don't write there)
; note: - pointer to buffer becomes invalid after next operation with stream
;       - on OF=0, do not read behind count bytes
;       - on OF=1, do not read buffer behind number of bytes returned in readlen variable.
;         If you passed 0 as readlen, don't read buffer at all
;       - this is not asynchronous read. This is just simple non-blocking read, it returns data
;         that is available immediately. It doesn't wait for data or EOF, like stream.read
; error: ERR_ZERO_SIZE           - count=0
;        ERR_ZERO_VALUE          - readlen=0
;        ERR_INVALID_HANDLE      - not a stream handle
;        ERR_HANDLE_NO_BUFFER    - stream doesn't have read buffer associated
;        ERR_HANDLE_NOT_READABLE - stream is not readable
;        ERR_BUFFER_FULL         - handle is associated with static buffer, with length smaller than count
;                                  on this error, nothing is read from stream
;        ERR_HANDLE_CORRUPT      - stream is corrupt, you are overwriting stream object somewhere,
;                                  or it's error in FASMLIB (unlikely)
;        ERR_FASMLIB             - internal FASMLIB bug! Please report.
;        +mem.realloc
;        +any error returned by stream readnb method
;===============================================================================================
proc stream.peeknb, stream, count, readlen
	push	ebx ecx edi
	pushf

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx+STREAM.sign], 'STRM'
	jne	.error_not_stream

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;error if readlen is zero
	cmp	[readlen], 0
	je	.error_readlen_zero

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if stream is not readable
	cmp	[ebx+STREAM.read], 0
	je	.error_stream_not_readable

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;use peeknb method, if provided
	cmp	[ebx+STREAM.rbuf.cnt], 0
	ja	@f
	mov	eax, [ebx+STREAM.peeknb]
	cmp	eax, 0
	je	@f
	libcall eax, [stream], [count], [readlen]
	jc	.rc ;TODO - we should handle ERR_BUFFER_FULL, and try to buffer at this level, in that case
	jo	.ro
	jmp	.rno
	@@:

	;ECX = count
	mov	ecx, [count]

	;if readnb is not supported, just return what we have on buffer
	cmp	[ebx+STREAM.readnb], 0
	je	.buffer_filled

	;error if we doesn't have any buffer
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.error_no_buffer

	;error if we have static buffer smaller than count
	cmp	[ebx+STREAM.rbuf.type], 1
	jne	@f
	cmp	ecx, [ebx+STREAM.rbuf.size]
	ja	.error_buffer_small
	@@:

	;if we have dynamic buffer, not yet allocated, then allocate it
	cmp	[ebx+STREAM.rbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.rbuf], 0
	jne	@f
	libcall mem.alloc, ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx
	mov	[ebx+STREAM.rbuf.pos], 0
	mov	[ebx+STREAM.rbuf.cnt], 0
	jmp	.buffer_long_enough
	@@:

	;see if buffer is big enough for count (without moving needed)
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	ecx, eax
	jbe	.buffer_long_enough

	;move data to beginning of buffer
	mov	eax, [ebx+STREAM.rbuf.pos]
	add	eax, [ebx+STREAM.rbuf]
	libcall mem.copy, [ebx+STREAM.rbuf], eax, [ebx+STREAM.rbuf.cnt]
	jc	.rc
	mov	[ebx+STREAM.rbuf.pos], 0

	;see if buffer is long enough now
	cmp	ecx, [ebx+STREAM.rbuf.size]
	jbe	.buffer_long_enough

	;if not, reallocate buffer to needed size
	libcall mem.realloc, [ebx+STREAM.rbuf], ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx

.buffer_long_enough:

	;check if there is already enough bytes in buffer
	cmp	ecx, [ebx+STREAM.rbuf.cnt]
	jbe	.buffer_filled

	;check if there is EOF marked in buffer
	cmp	[ebx+STREAM.rbuf.eof],1
	je	.buffer_filled

	;try to read remaining bytes with readnb, if provided
	mov	eax, [ebx + STREAM.readnb]
	cmp	eax, 0
	je	@f
	mov	edi, [ebx+STREAM.rbuf]
	add	edi, [ebx+STREAM.rbuf.pos]
	add	edi, [ebx+STREAM.rbuf.cnt]
	sub	ecx, [ebx+STREAM.rbuf.cnt]
	libcall eax, [stream], edi, ecx
	jc	.rc
	@@:

	;adjust buffer info (count, EOF flag)
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	add	[ebx+STREAM.rbuf.cnt], eax

.buffer_filled:

	;readlen = min(count, readbuf.cnt)
	mov	ecx, [count]
	mov	eax, [ebx+STREAM.rbuf.cnt]
	cmp	ecx, eax
	cmova	ecx, eax
	mov	eax, [readlen]
	mov	[eax], ecx

	;eax = readbuf
	mov	eax, [ebx + STREAM.rbuf]
	add	eax, [ebx + STREAM.rbuf.pos]

	;return OF if count<readbuf.cnt and rbuf.eof is set
	cmp	ecx, [count]
	je	@f
	ja	.error_internal
	cmp	[ebx + STREAM.rbuf.eof], 1
	je	.ro
	ja	.error_internal  ;assert rbuf.eof <= 1
	@@:

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc


.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_readlen_zero:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc

.error_no_buffer:
	mov	eax, ERR_HANDLE_NO_BUFFER
	jmp	.rc

.error_buffer_small:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc
endp



;===============================================================================================
; stream.skip
; desc: skip data on stream input. It's like reading them "nowhere" from stream.
; args: stream  - handle of stream from which to skip
;       count   - number of bytes to skip (can be 0)
; ret: CF set on error, otherwise
;      OF set if EOF was reached (less than count bytes skipped), and
;      EAX = number of bytes skipped (on OF=0, always EAX=count)
; error: ERR_INVALID_HANDLE      - not a stream handle
;        ERR_HANDLE_NOT_READABLE - stream is not readable
;        +any error returned by stream reading procedure
;===============================================================================================
proc stream.skip stream, count
	push	ebx ecx edx edi
	pushf

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx+STREAM.sign], 'STRM'
	jne	.error_not_stream

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;error if stream is not readable
	cmp	[ebx+STREAM.read], 0
	je	.error_stream_not_readable

	;ECX = bytes to skip
	mov	ecx, [count]

	;if it's zero, just return
	cmp	ecx, 0
	cmove	eax, ecx
	je	.rno

	;if we don't have buffering enabled, go directly to method call
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.call_method

	;if we don't have anything on buffer, go to method call
	cmp	[ebx+STREAM.rbuf.cnt], 0
	je	.call_method

	;EDX = bytes to remove from buffer = min(count, bytes_on_buffer)
	mov	edx, ecx
	mov	eax, [ebx+STREAM.rbuf.cnt]
	cmp	edx, eax
	cmova	edx, eax

	;remove them from buffer
	add	[ebx+STREAM.rbuf.pos], edx
	sub	[ebx+STREAM.rbuf.cnt], edx
	jnz	@f
	mov	[ebx+STREAM.rbuf.pos], 0 ;we cleared entire buffer, so move pos to beginning
	@@:

	;end, if enough data removed
	sub	ecx, edx
	jne	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;end, if EOF was stored on buffer
	cmp	[ebx+STREAM.rbuf.eof], 1
	jne	@f
	mov	eax, [count]
	sub	eax, ecx
	jmp	.ro
	@@:

	;otherwise, continue removing rest with skip or read

.call_method:

	;eax = skip method, use read if unsupported
	mov	eax, [ebx+STREAM.skip]
	cmp	eax, 0
	je	.use_read_method
	libcall eax, [stream], ecx

	;done if we skipped enough bytes without reaching EOF
	jo	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;if we reached eof, set EOF flag and return number of bytes read
	mov	[ebx+STREAM.rbuf.eof], 1
	mov	eax, [count]
	sub	eax, ecx
	jmp	.ro


.use_read_method:

	;EDI = intermediate buffer
	sub	esp, 512
	mov	edi, esp

.read_loop:
	;size of data to read
	mov	edx, 512
	cmp	edx, ecx
	cmova	edx, ecx

	;read into buffer
	libcall [ebx+STREAM.read], [stream], edi, edx
	jc	.error_read

	;return if EOF was reached
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	add	esp, 512
	sub	ecx, eax
	mov	eax, [count]
	sub	eax, ecx
	jmp	.ro
	@@:

	;loop if we didn't skip enogh
	sub	ecx, eax
	jne	.read_loop
	add	esp, 512

	;we've skipped enough. return EAX=count, OF=0
	mov	eax, [count]

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	cmp	[ebx+STREAM.rbuf.eof], 1 ;just some assertion to catch internal errors
	jne	.error_corrupt_stream
	jmp	.rnc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc

.error_read:
	;remove buffer from stack
	add	esp, 512
	jmp	.rc
endp



;===============================================================================================
; stream.close
; desc: closes stream
; args: stream - handle of stream which we are closing
; ret: CF set on error
; note: this is in fact just a helper proc which hides OOP aspect of streams ;)
; error: ERR_INVALID_HANDLE - "stream" is not a valid stream handle
;        ERR_HANDLE_CORRUPT - handle is corrupt. probably you have overwriten structure in
;                             memory. This might also be a FASMLIB bug.
;        +any error returned by stream closing procedure
;===============================================================================================
proc stream.close, stream
	push	ebx

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_not_stream

	;EAX = close method, skip if not present
	mov	eax, [stream]
	mov	eax, [eax + STREAM.close]
	cmp	eax, 0
	je	@f
	libcall eax, [stream]
	jc	.error_method_failed
	@@:

	;destroy buffer if it is dynamic and allocated
	cmp	[ebx+STREAM.rbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.rbuf], 0
	je	@f
	libcall mem.free, [ebx+STREAM.rbuf]
	jc	.error_buffer_freeing_failed
	@@:

	;destroy object
	cmp	[ebx+STREAM.allocated], 1
	jne	@f
	libcall mem.free, ebx
	jc	.rc
	@@:

	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_method_failed:
	push	eax
	libcall mem.free, [stream]
	pop	eax
	jmp	.rc

.error_buffer_freeing_failed:
	push	eax
	libcall mem.free, [stream]
	pop	eax

	;translate ERR_INVALID_HANDLE to ERR_HANDLE_CORRUPT (pointer to buffer was invalid)
	cmp	eax, ERR_INVALID_HANDLE
	jne	@f
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc
	@@:

	jmp	.rc

endp



;by vid