;FASM STANDARD LIBRARY
;Win32 filesystem routines

;contains:
; proc  file.open       name, mode
; proc  file.create     name, mode
; proc  file.close      handle
; proc  file.seek       handle, dist, method
; proc  file.read       handle, buffer, count
; proc  file.write      handle, buffer, count



;============================================================================
;file.open
;desc: opens file
;args: name - pointer to null terminated string containing file name
;      mode - 0: read, 1: write, 2:r/w
;ret: CF on error, otherwise
;     EAX = handle to file
;note: don't forget to file.close returned handle!
;error: ERR_FILE_NOT_FOUND - file not found
;       ERR_CANT_OPEN_FILE - file exists, but we failed to open it
;       ERR_INVALID_MODE - invalid mode
;       ERR_UNKNOWN
;============================================================================
proc file.open name, mode

	push ecx edx

	;push some args
	push	0		;hTemplateFile
	push	0		;dwFlagsAndAttributes
	push	3		;dwCreationDistributiuon - OPEN_EXISTING
	push	0		;pointer to SECURITY_ATTRIBUTES

	;push share mode depending on [.mode]
	cmp	[mode], 0      ;read
	jne	@f
	push	1		;FILE_SHARE_READ
	jmp	.sharemode_done
@@:	push	0
.sharemode_done:


	;push proper mode flag depending on [.mode]
	cmp	[mode], 0
	jnz	@f
	push	80000000h	;GENERIC_READ
	jmp	.modedone
@@:
	cmp	[mode], 1
	jnz	@f
	push	40000000h	;GENERIC_WRITE
	jmp	.modedone
@@:
	cmp	[mode], 2
	jnz	@f
	push	0C0000000h	;GENERIC_READ | GENERIC_WRITE
	jmp	.modedone
@@:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
.modedone:

	;push file name
	push	[name]

	;call it
	call	[CreateFileA]
	cmp	eax,-1
	je	.error_CreateFile

	;return withot error
.rnc:	clc

	;return
.r:	pop	edx ecx
	ret

.rc:	stc
	jmp	.r

	;CreateFile returned with error
.error_CreateFile:
	call	[GetLastError]

	;catch file not found errors
	push	ERR_FILE_NOT_FOUND
	cmp	eax, 2			;ERROR_FILE_NOT_FOUND
	je	@f
	cmp	eax, 3			;ERROR_PATH_NOT_FOUND
	je	@f
	cmp	eax, 15 		;ERROR_INVALID_DRIVE
	je	@f
	cmp	eax, 123		;ERROR_INVALID_NAME
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	;catch can't open file errors
	push	ERR_CANT_OPEN_FILE
	cmp	eax, 4			;ERROR_TOO_MANY_OPEN_FILES
	je	@f
	cmp	eax, 5			;ERROR_ACCESS_DENIED  (also returned when trying to "open" directory
	je	@f
	cmp	eax, 19 		;ERROR_WRITE_PROTECT
	je	@f
	cmp	eax, 30 		;ERROR_READ_FAULT
	je	@f
	cmp	eax, 32 		;ERROR_SHARING_VIOLATION
	je	@f
	cmp	eax, 33 		;ERROR_LOCK_VIOLATION
	je	@f
	cmp	eax, 36 		;ERROR_SHARING_BUFFER_EXCEEDED
	je	@f
	cmp	eax, 110		;ERROR_OPEN_FAILED
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;============================================================================
;file.create
;desc: creates new file
;args: name - pointer to null terminated string caontaining file name
;      mode - bit0: 1 if overwrite when file exists
;             bit1: 1 if you want read access too
;             all other bits 0
;ret: CF on error, otherwise
;     EAX = handle to file
;note: don't forget to close file handle with file.close!
;error: ERR_FILE_EXISTS        - file already exists
;       ERR_CANT_OPEN_FILE     - file can't be created
;       ERR_INVALID_MODE       - if mode is invalid
;       ERR_UNKNOWN
;============================================================================
proc file.create name, mode
	push ecx edx

	;check if mode is <4
	cmp	[mode], 4
	jb	@f
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
@@:

	;push some args
	push	0 ;hTemplateFile
	push	0 ;dwFlagsAndAttributes

	;overwrite existing? depending on [mode] bit 0
	test	[mode],1b
	jz	@f
	push	2	   ;CREATE_ALWAYS
	jmp	.create_mode_done
@@:	push	1	   ;CREATE_NEW
.create_mode_done:

	;push some more junk
	push	0 ;pointer to SECURITY_ATTRIBUTES
	push	0 ;dwShareMode

	;push proper access flag depending on [mode] bit 1
	test	[mode],10b
	jz	@f
	push	0C0000000h ;GENERIC_READ | GENERIC_WRITE
	jmp	.access_mode_done
@@:	push	40000000h ;GENERIC_WRITE
.access_mode_done:

	;push file name
	push	[name]

	;call CreateFileA
	call	[CreateFileA]
	cmp	eax, -1
	je	.error_CreateFile

	;no error occured, return system handle
	clc

	;return
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

	;CreateFile returned with error
.error_CreateFile:
	call	[GetLastError]

	;catch already-exists
	push	ERR_FILE_EXISTS
	cmp	eax, 80 		;ERROR_FILE_EXISTS
	je	@f
	cmp	eax, 183		;ERROR_ALREADY_EXISTS
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	;rest can be classified as can't create
	mov	eax, ERR_CANT_OPEN_FILE
	jmp	.rc
endp



;============================================================================
;file.close
;desc: closes file handle
;args: handle - file handle (returned by file.open)
;ret: CF set on error
;error: ERR_INVALID_HANDLE - wrong handle
;       ERR_UNKNOWN
;============================================================================
proc file.close handle
	push	ecx edx

	push	[handle]
	call	[CloseHandle]

	or	eax,eax
	jz	.error_CloseHandle
	clc


.r:	pop	edx ecx
	ret

.rc:	stc
	jmp .r

.error_CloseHandle:
	call	[GetLastError]

	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	jne	@f
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
@@:
	mov	eax, ERR_UNKNOWN
	jmp	.rc

endp



;============================================================================
;file.seek
;desc: seek position in file
;args: handle - file handle (returned by file.open)
;      dist - signed distance to move
;      mode - how to seek (0: from beginning, 1:relative to current, 2: from end)
;ret: CF set on error, otherwise
;     EAX = new file position (from beginning)
;note: - you can use this to get current position
;          libcall file.seek, <handle>, 0, 1
;      - you can use this to get size of opened file:
;          libcall file.seek, <handle>, 0, 2
;        but note that this also changes position in file to end
;      - you can safely use this call only on files smaller than 2GB
;error: ERR_INVALID_HANDLE - invalid file handle
;       ERR_OUT_OF_RANGE   - dist<0 when method=0, or seek to negative file offset
;       ERR_INVALID_MODE   - method is not 0, 1, or 2
;       ERR_UNKNOWN
;============================================================================
proc file.seek handle, dist, mode
	push ecx edx

	;check if mode is valid
	cmp	[mode], 2
	jbe	@f
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
@@:

	;check if [.dist] is positive if [.method] is 0  (windoze would check this, but this is nicer)
	cmp	[mode], 0
	jne	@f
	cmp	[dist], 0
	jge	@f
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
@@:

	;call API
	push	[mode]			;windows uses same numbers as we do
	push	0			;ptr to high order dword of distance to move, 0 means ignore
	push	[dist]
	push	[handle]
	call	[SetFilePointer]
	cmp	eax,-1
	je	.error_SetFilePointer

	clc

.r:	pop	edx ecx
	ret

.rc:	stc
	jmp	.r

.error_SetFilePointer:
	call	[GetLastError]

	;handle invalid type of handle :)
	push	ERR_INVALID_HANDLE
	cmp	eax, 6		;ERROR_INVALID_HANDLE
	je	@f
	cmp	eax, 132	;ERROR_SEEK_ON_DEVICE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	;handle negative seek
	push	ERR_OUT_OF_RANGE
	cmp	eax, 131	;ERROR_NEGATIVE_SEEK
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;============================================================================
;file.read
;desc: read data from current position in file
;args: handle - file handle (returned by file.open)
;      buffer - pointer to buffer where to store data
;      count - number of bytes to read
;ret: CF set on error, otherwise
;     OF set on read behind EOF (less that <count> bytes were readen)
;     EAX = number of bytes readen
;error: ERR_INVALID_HANDLE - invalid handle, or handle is write-only
;       ERR_ZERO_SIZE      - count is zero
;       ERR_UNKNOWN
;============================================================================
proc file.read handle, buffer, count
  locals
    .bytesread dd ?
  endl
	push	ecx edx
	pushf

	;check if count>0
	cmp	[count], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:
	;call it
	push	0 ;lpOverlapped
	lea	eax,[.bytesread]
	push	eax
	push	[count]
	push	[buffer]
	push	[handle]
	call	[ReadFile]

	;check for error
	or	eax,eax
	je	.error_ReadFile

	;no error, clear CF
	btr	dword [esp], 0 ;clear CF

	;set OF if bytes readen != bytes to read
	btr	dword [esp], 11 ;clear OF
	mov	eax,[.bytesread]
	cmp	eax,[count]
	je	@f
	bts	dword [esp], 11 ;set OF
@@:

	;return number of bytes readen in EAX
	mov	eax,[.bytesread]

.r:	popf
	pop	edx ecx
	ret

	;error occured
.rc:	bts	dword [esp], 0 ;set CF
	jmp	.r

.error_ReadFile:
	call	[GetLastError]

	push	ERR_INVALID_HANDLE
	cmp	eax, 5 ;ERROR_ACCESS_DENIED    - calling file.read with write-only handle can be considered invalid argument
	je	@f
	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	je	@f
	mov	[esp], eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;======================================================================
;file.write
;desc: writes data to current position in file
;args: handle - file handle (returned by file.open)
;      buffer - pointer to buffer from where to read data
;      count - number of bytes to read
;ret: CF set on error
;error: ERR_INVALID_HANDLE - invalid handle
;       ERR_ZERO_SIZE      - count is zero
;       ERR_UNKNOWN
;======================================================================
proc file.write handle, buffer, count
  locals
    .byteswritten dd ?
  endl

	push	ecx edx

	;check if count>0
	cmp	[count], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:

	;call it
	push	0 ;lpOverlapped
	lea	eax,[.byteswritten]
	push	eax
	push	[count]
	push	[buffer]
	push	[handle]
	call	[WriteFile]

	;check for error
	or	eax,eax
	je	.error_WriteFile

	;check bytes written
	mov	eax,[.byteswritten]
	cmp	eax,[count]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
@@:

	;no error
	clc

.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_WriteFile:
	call	[GetLastError]

	;TODO - ERROR_ACCESS_DENIED

	;catch invalid handle
	push	ERR_INVALID_HANDLE
	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp


;by vid