
;FASM STANDARD LIBRARY
;Win32 memory routines

;contains:
; proc  mem.init
; proc  mem.uninit
; proc  mem.alloc       size
; proc  mem.free        blockptr
; proc  mem.realloc     blockptr,newsize
; proc  mem.size        blockptr

idata {
  mem.initialized dd 0		;0 = not initialized, 1=initialized
}

udata {
  mem.heap	  dd ?		       ;heap handle
}

;============================================================================
; mem.init
; desc: initalizes this module
; args: none
; ret: CF set on error
; note: initialization is required for allocation
; error: ERR_UNKNOWN - platform-specific error
;============================================================================
proc mem.init
	push	ecx edx

	cmp	[mem.initialized],0
	jne	.rnc

	;get heap pointer
	call	[GetProcessHeap]

	or	eax, eax
	jz	.error_GetProcessHeap

	mov	[mem.heap], eax
	mov	[mem.initialized], 1

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_GetProcessHeap:
	mov eax, ERR_UNKNOWN
	jmp .rc
endp



;============================================================================
; mem.uninit
; desc: uninitalizes this module
; args: none
; ret: CF set on error
; note: ignored if module is not initialized
; error: ERR_MODULE_NOT_INITIALIZED - Module is not initialized yet
;        ERR_UNKNOWN                - Platform-specific error
;============================================================================
proc mem.uninit
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;store state
	mov	[mem.initialized],0

.rnc:	clc
.r:	pop	edx ecx
	ret

.error_HeapDestroy:
	mov	eax, ERR_UNKNOWN
	stc
	jmp	.r

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	stc
	jmp	.r
endp


;============================================================================
; mem.alloc
; desc: allocates <size> bytes of memory
; args: size - size of memory to allocate
; ret: CF set on error, otherwise
;      eax = pointer to beginning of allocated memory block
; note: - returned pointer must be always >= 10000h
;       - contents of allocated memory are unknown
;         (use alloc0 to have block zeroed)
;       - don't forget to mem.free allocated memory!
;       - maximal allowed size if 7FFFFFFFh, to catch errors with negative value
; error: ERR_OUT_OF_MEM   - out of memory
;        ERR_ZERO_SIZE    - size=0
;        ERR_OUT_OF_RANGE - size >= 80000000h
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;============================================================================
proc mem.alloc size
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;error if signed size<=0
	cmp	[size], 0
	je	.error_size_zero
	jl	.error_size_too_big

	;size += header
	mov	ecx, [size]
	add	ecx, 8

	;call HeapAlloc
	push	ecx
	push	dword 0 ;no flags
	push	[mem.heap]
	call	[HeapAlloc]
	or	eax,eax
	jz	.error_HeapAlloc

	;fill header
	mov	dword [eax], 'RnR'  ;sign
	mov	ecx, [size]
	mov	dword [eax+4], ecx  ;size (not including header)

	;return pointer behind header
	add	eax, 8

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapAlloc:
	call	[GetLastError]

	;catch out-of-memory
	push	ERR_OUT_OF_MEM
	cmp	eax, 8			;ERROR_NOT_ENOUGH_MEMORY
	je	@f
	cmp	eax, 14 		;ERROR_OUTOFMEMORY
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_size_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp



;=============================================================================
; mem.free
; desc: deallocates memory block
; args: blockptr - pointer to memory block returned by mem.alloc
; ret: CF set on error
; note: after error, consider block as freed, don't use it anymore
; error: ERR_INVALID_POINTER - if blockptr isn't valid block pointer
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;=============================================================================
proc mem.free blockptr
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;EAX = memory block (before header)
	mov	eax, [blockptr]
	sub	eax, 8

	;check if it is our memory block
	cmp	dword [eax], 'RnR'
	je	@f
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
@@:

	;destroy 'RnR' mark - because memory in heap remains accessible even after it is freed
	mov	dword [eax], 0

	;call HeapFree
	push	eax		;pointer to beginning of block (including header)
	push	0		;no flags
	push	[mem.heap]
	call	[HeapFree]
	or	eax,eax
	jz	.error_HeapFree

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapFree:
	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

endp



;=============================================================================
; mem.realloc
; desc: reallocates memory block
; args: blockptr - pointer to memory block returned by mem.alloc
;       newsize - new size of memory block
; ret: CF set on error, orherwise
;      eax = pointer to reallocated memory block
; note: returned pointer must be always >= 10000h
;       - data in block are is modified (only data at the end of block is
;         lost if newsize < previous size)
;       - if block is enlarging, then contents of added memory are unknown
;         (use realloc0 to have it zeroed)
;       - block can move to other place in memory, so previous pointers into
;         block become invalid (so rather use indexes in block than pointers)
;       - maximal allowed size if 7FFFFFFFh, to catch errors with negative value
;       - if you had pointer stored in variable, don't forget to restore it:
;           libcall mem.realloc, [blockptr], [newsize]
;           jc error
;           mov [blockptr], eax     ;without this, it's hard-to-find bug
; error: ERR_OUT_OF_MEM     - windoze is gonna' resize swap file for a minute and crash your app
;        ERR_INVALID_POINTER - if blockptr isn't valid block pointer
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;=============================================================================
proc mem.realloc blockptr, newsize
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;error if size=0
	cmp	[newsize], 0
	je	.error_newsize_zero
	jl	.error_newsize_too_big

	;EAX = memory block - header
	mov	eax, [blockptr]
	sub	eax, 8

	;check if it is our memory block
	cmp	dword [eax], 'RnR'
	je	@f
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
@@:

	;call HeapReAlloc
	mov	ecx, [newsize]
	add	ecx, 8			;size of block + header
	push	ecx
	push	eax			;pointer to old memory block before header
	push	0			;no flags
	push	[mem.heap]		;our heap handle
	call	[HeapReAlloc]
	or	eax,eax
	jz	.error_HeapRealloc

	;fill new header (needed when block has moved)
	mov	dword [eax], 'RnR'	;sign
	mov	ecx, [newsize]
	mov	[eax+4], ecx		;size (not including header)

	;return pointer behind header
	add	eax, 8

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapRealloc:
	call [GetLastError]

	;catch out-of-memory
	push	ERR_OUT_OF_MEM
	cmp	eax, 8			;ERROR_NOT_ENOUGH_MEMORY
	je	@f
	cmp	eax, 14 		;ERROR_OUTOFMEMORY
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_newsize_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp



;=============================================================================
; mem.size
; desc: returns size of memory block
; args: blockptr - pointer to memory block returned by mem.alloc
; ret: CF set on error, orherwise
;      eax = size of memory block
; error: ERR_INVALID_POINTER - blockptr isn't a valid block pointer
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;=============================================================================
proc mem.size blockptr
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	je	.error_not_initialized

	;EAX = memory block
	mov	eax, [blockptr]

	;check if it is our memory block
	cmp	dword [eax-8], 'RnR'
	je	@f
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
@@:

	;return size of block
	mov	eax, [eax-4]
	clc

	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc
endp


;by vid