;FASM STANDARD LIBRARY
;Win32 stream I/O routines

;contains:
; proc stream.init
; proc stream.uninit

;internal:
; proc stream.write_stdout       stream, buffer, size
; proc stream.write_stderr       stream, buffer, size
; proc stream.read_stdin         stream, buffer, size


;TODO - add stream.readnb_stdin for case when stdin is redirected from file


;predefined streams:
idata {
  stream.initialized dd 0

  stream.stdout STREAM\
		'STDO', 	       \;type
		0,		       \;read
		0,		       \;peek
		0,		       \;skip
		stream.write_stdout,   \;write
		0,		       \;readnb
		0,		       \;peeknb
		0,		       \;skipnb
		0,		       \;flush
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0,		       \;readbufsize    = not yet allocated
		0, 0, 0 		;no write buf

  stream.stderr STREAM\
		'STDE', 	       \;type
		0,		       \;read
		0,		       \;peek
		0,		       \;skip
		stream.write_stderr,   \;write
		0,		       \;readnb
		0,		       \;peeknb
		0,		       \;skipnb
		0,		       \;flush
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0,		       \;readbufsize    = not yet allocated
		0, 0, 0 		;no write buf

  stream.stdin	STREAM\
		'STDI', 	       \;type
		stream.read_stdin,     \;read
		0,		       \;peek
		0,		       \;skip
		0,		       \;write
		0,		       \;readnb - TODO: add with O_NONBLOCK
		0,		       \;peeknb - TODO: add with O_NONBLOCK
		0,		       \;skipnb
		0,		       \;flush
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0,		       \;readbufsize    = not yet allocated
		0, 0, 0 		;no write buf
}


proc stream.init
	clc
	ret
endp


proc stream.uninit

	;to free read buffer
	libcall stream.close, stream.stdin
	jc	.rc

.rnc:	clc
.r:	ret
.rc:	stc
endp

;===============================================================================================
; stream.write_stdout
; desc: writes (blocking) to system standard output
; args: buffer - pointer to buffer which holds data
;       size - size of buffer (in bytes)
; ret: CF set on error
; note: this procedure is not to be called directly. it's used by stream operations on stream.stdout
; error: ERR_INVALID_HANDLE - invalid stream handle
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;===============================================================================================
proc stream.write_stdout, stream, buffer, size
local written dd ?

	push	ecx edx

	;check if size=0
	cmp	[size], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:

	push	-11 ;STD_OUTPUT_HANDLE
	call	[GetStdHandle]
	mov	ecx, eax

	;call WriteFile
	push	0			;reserved
	lea	eax,[written]
	push	eax			;number of chars written
	push	[size]
	push	[buffer]
	push	ecx
	call	[WriteFile]
	or	eax,eax
	jz	.error_WriteFile

	;check if all bytes were written
	mov	eax,[written]
	cmp	eax,[size]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
@@:

	;return no unwritten bytes
	mov	eax, 0

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	ret

.error_WriteFile:
	call	[GetLastError]

	;catch invalid handle
	push	ERR_INVALID_HANDLE
	cmp	eax, 5 ;ERROR_ACCESS_DENIED    - calling file.write with read-only handle can be considered invalid argument
	je	@f
	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;===============================================================================================
; stream.write_stderr
; desc: writes (blocking) to system standard error
; args: buffer - pointer to buffer which holds data
;       size - size of buffer (in bytes)
; ret: CF set on error
; note: this procedure is not to be called directly. it's used by stream operations on stream.stderr
; error: ERR_INVALID_HANDLE - invalid stream handle
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;===============================================================================================
proc stream.write_stderr, stream, buffer, size
local written dd ?

	push	ecx edx

	;check if size=0
	cmp	[size], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
@@:

	push	-12 ;STD_ERROR_HANDLE
	call	[GetStdHandle]
	mov	ecx, eax

	;call WriteFile
	push	0			;reserved
	lea	eax,[written]
	push	eax			;number of chars written
	push	[size]
	push	[buffer]
	push	ecx
	call	[WriteFile]
	or	eax,eax
	jz	.error_WriteFile

	;check if all bytes were written
	mov	eax,[written]
	cmp	eax,[size]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
@@:

	;return no unwritten bytes
	xor	eax, eax

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	ret

.error_WriteFile:
	call	[GetLastError]

	;catch invalid handle
	push	ERR_INVALID_HANDLE
	cmp	eax, 5 ;ERROR_ACCESS_DENIED    - calling file.write with read-only handle can be considered invalid argument
	je	@f
	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp


;===============================================================================================
; stream.read_stdin
; desc: reads from system standard input. blocks if there isn't enough data typed already.
;       this is used by "libcall stream.read, stream.stdin, ...". Don't call directly unless you
;       really have to
; args: stream - pointer to stream
;       buffer - pointer to buffer which holds data
;       count  - number of bytes to read
; ret: CF set on error, otherwise
;      OF set on EOF, and
;      EAX = number of bytes read  (if OF=0, then EAX=count)
; error: ERR_INVALID_HANDLE - called on other stream than stream.stdin
;        ERR_ZERO_SIZE      - if size=0
;        ERR_FASMLIB        - bug in FASMLIB
	 ;        ERR_UNKNOWN
;===============================================================================================
proc stream.read_stdin, stream, buffer, count
local readen dd ?
	push	ebx ecx edx edi
	pushf

	;check if this is STDIN
	mov	eax, [stream]
	cmp	[eax + STREAM.type], 'STDI'
	jne	.error_invalid_stream

	;error if count=0
	cmp	[count], 0
	je	.error_count_zero

	;EBX = size
	mov	ebx, [count]

	;EDI = buffer
	mov	edi, [buffer]

.read:
	push	-10 ;STD_INPUT_HANDLE
	call	[GetStdHandle]
	mov	edx, eax

	;call ReadFile
	;function returns after reading one line, which can be less than size we want
	push	0			;reserved
	lea	eax,[readen]
	push	eax			;pointer to variable to hold number of chars readen
	push	ebx			;number of bytes to read
	push	edi			;pointer to buffer
	push	edx			;stdout handle
	call	[ReadFile]
	or	eax,eax
	jz	.error_ReadFile

	;if return value is non-zero, and bytes readen is zero, then we are on eof
	cmp	[readen], 0
	je	.done

	;end check if all bytes were readen
	sub	ebx, [readen]
	je	.done

	;error if we read more than we wanted (???)
	jb	.error_readen_more_than_wanted

	;continue reading next line
	add	edi, [readen]
	jmp	.read

.done:
	;EAX = number of bytes readen
	mov	eax, [count]
	sub	eax, ebx

	;set OF if we didn't read all bytes (EOF)
	cmp	ebx, 0
	ja	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_ReadFile:
	call	[GetLastError]

	;catch invalid handle
	push	ERR_FASMLIB
	cmp	eax, 6 ;ERROR_INVALID_HANDLE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_readen_more_than_wanted:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp



;by vid